module Corba(
	   ResultOrException(..), 
	   is_just, just, was_exception, corba_result_create,
	   TypeCode(..), corba_tc_encode, corba_tc_decode,
	   Any(..), corba_any_encode, corba_any_decode,
	   CompletionStatus(..),
	   CorbaException(..), decode_corba_exception, corba_exception_encode,
	   CorbaArg(..), corba_arg_get_value,
	   corba_invoke,
	   CORBA_Void, corba_Void__tc, corba_Void__decode, corba_Void__encode,
	   CORBA_Short, corba_Short__tc, corba_Short__decode, corba_Short__encode,
	   CORBA_Long, corba_Long__tc, corba_Long__decode, corba_Long__encode,
	   CORBA_UShort, corba_UShort__tc, corba_UShort__decode, corba_UShort__encode,
	   CORBA_ULong, corba_ULong__tc, corba_ULong__decode, corba_ULong__encode,
	   CORBA_Float, corba_Float__tc, corba_Float__decode, corba_Float__encode,
	   CORBA_Double, corba_Double__tc, corba_Double__decode, corba_Double__encode,
	   CORBA_Boolean, corba_Boolean__tc, corba_Boolean__decode, corba_Boolean__encode,
	   CORBA_Char, corba_Char__tc, corba_Char__decode, corba_Char__encode,
	   CORBA_Octet, corba_Octet__tc, corba_Octet__decode, corba_Octet__encode,
	   CORBA_Any, corba_Any__tc, corba_Any__decode, corba_Any__encode,
	   CORBA_TypeCode, corba_TypeCode__tc, corba_TypeCode__decode, corba_TypeCode__encode,
	   CORBA_String, corba_String__tc, corba_String__decode, corba_String__encode,
	   CORBA_Object, corba_Object__tc, corba_Object__decode, corba_Object__encode,
	   corba_Object_is_a, corba_Object__create, corba_Object__nil,
	   CORBA_ORB, corba_ORB_init, corba_ORB_bind, corba_ORB_string_to_object,
	   corba_ORB_object_to_string, corba_ORB_resolve_initial_references,
	   writeIORToFile, readIORFromFile,
	   CORBA_Object__widen_class(..), 

	   CORBA_ServerRequest, corba_ServerRequest_operation, 
	   corba_ServerRequest_set_arg_types, corba_ServerRequest_get_arg,
	   corba_ServerRequest_set_arg, corba_ServerRequest_set_result,
	   corba_ServerRequest_set_exception,

	   PSt, IOSt, ServerInitFunc, ServantDispatchFun, CORBA_Servant_Wrapper(..),
	   noInit, corba_Server_run, corba_Server_exit, corba_Server_add_servant,
	   corba_Server_get_current, corba_Server_get_orb,
	   corba_enum_decode, corba_objref_decode, corba_objref_narrow,
	   corba_struct_decode, corba_exception_decode, corba_seq_decode,
	   corba_array_decode, corba_union_decode, corba_get_objref,
	   narrow_objref, corba_invoke_simple

) where

import CorbaWrappers
import IO
import IOExts
import Foreign
import IORef

--------------------------------
--       UTILITIES
--------------------------------

is_just :: (Maybe a) -> Bool
is_just (Just _)	= True
is_just  _			= False

just :: (Maybe a) -> a
just (Just x)		= x

was_exception :: CorbaException -> Bool
was_exception NoException 	= False
was_exception _ 			= True

data ResultOrException a b = NoResult | Result a | Exception b

-- If there was an exception, the result is CorbaVoid
-- Luckily, lazy evaluation prevents the system from trying to decode
-- it into another CORBA type...
corba_result_create :: a -> CorbaException -> (ResultOrException a CorbaException)
corba_result_create result exception
	| was_exception exception
	= Exception exception
	| otherwise
	= Result result

--------------------------------
--  HANDLING CORBA SEQUENCES
--------------------------------

-- Higher-order function for creating CORBA Sequences from a Haskell list

type SeqCreateFct seq_type			=	Int -> IO seq_type
type SeqSetFct seq_type member_type	=	seq_type -> Int -> member_type -> IO ()

encode_corba_seq :: (SeqCreateFct seq_type)
				 -> (SeqSetFct seq_type member_type)
				 -> [member_type] 
				 -> IO seq_type
encode_corba_seq create_fun set_fun l
	= do
	  seq <- create_fun (length l)
	  sequence_ $ map (\(i, v) -> (set_fun seq i v)) (zip [0..] l)
	  return seq

type SeqGetFct seq_type member_type = seq_type -> Int -> IO member_type

-- Higher-order function for decoding CORBA Sequences into a Haskell list

decode_corba_seq :: (SeqGetFct seq_type member_type) -> seq_type -> Int
					-> IO [member_type]
decode_corba_seq get_fun seq len
	= sequence [ (get_fun seq i) | i <- [0..len-1] ]

--------------------------------
--       CStringArray
--------------------------------

c_string_array_encode :: [String] -> IO (CStringArrayPtr, Int)
c_string_array_encode l
	= do
	  arr <- encode_corba_seq c_string_array_create c_string_array_set l
	  return (arr, length l)

c_string_array_decode :: CStringArrayPtr -> Int -> IO [String]
c_string_array_decode arr len = decode_corba_seq c_string_array_get arr len

--------------------------------
--        TYPECODE
--------------------------------

-- HIGH LEVEL INTERFACE

data TypeCode = Tk_Null | 
				Tk_Void | 
				Tk_Short | 
				Tk_Long | 
				Tk_UShort |
				Tk_ULong | 
				Tk_Float | 
				Tk_Double | 
				Tk_Boolean | 
				Tk_Char |
				Tk_Octet | 
				Tk_Any | 
				Tk_String Int | 
				Tk_TypeCode | 
				Tk_ObjRef String String | 
				Tk_Struct String String [(String, TypeCode)] | 
			  	Tk_Sequence Int TypeCode |
			  	Tk_Union String String TypeCode [(String, Any, TypeCode)] |
				Tk_Enum String String [String] |
				Tk_Alias String String TypeCode |
				Tk_Array Int TypeCode |
				Tk_Exception String String [(String, TypeCode)] |
				Tk_Recursive String
  deriving Show

-- TODO:
instance Eq TypeCode where
	(==) tc1 tc2 = True

-- The 'repoids' parameter is used for handling recursive typecodes
corba_struct_member_seq_encode_with :: [(String, TypeCode)] -> [String] -> IO StructMemberSeqPtr
corba_struct_member_seq_encode_with members repoids
	= encode_corba_seq (corba_struct_member_seq_create) (add_member) members
	where
		add_member :: StructMemberSeqPtr -> Int -> (String, TypeCode) -> IO ()
		add_member seq index (name, tc)
		   = do
				tcPtr <- corba_tc_encode_with tc repoids
				corba_struct_member_seq_fill seq index name tcPtr

corba_union_member_seq_encode_with :: [(String, Any, TypeCode)] -> [String] -> IO UnionMemberSeqPtr
corba_union_member_seq_encode_with members repoids
	= encode_corba_seq (corba_union_member_seq_create) (add_member) members
	where
		add_member :: UnionMemberSeqPtr -> Int -> (String, Any, TypeCode) -> IO ()
		add_member seq index (name, label, tc)
			= do
				labelPtr <- corba_any_encode label
				typePtr	 <-	corba_tc_encode_with tc repoids
				corba_union_member_seq_fill seq index name labelPtr typePtr

corba_enum_member_seq_encode :: [String] -> IO EnumMemberSeqPtr
corba_enum_member_seq_encode members
	= encode_corba_seq (corba_enum_member_seq_create) (add_member) members
	where
		add_member seq index name = corba_enum_member_seq_fill seq index name

--
-- Handling recursive typecodes:
--   The 'encode' and 'decode' functions deals with recursive typecodes 
-- automatically, by inserting Tk_Recursive typecodes where appropriate.
-- This means the caller is free to pass infinite typecodes such as
-- tc = (Tk_Struct "x" "x" [("bla", (Tk_Sequence 0 tc))])
--

corba_tc_encode :: TypeCode -> IO TypeCodePtr
corba_tc_encode tc = corba_tc_encode_with tc []

-- The second argument contains the repoids of IDL structures and unions
-- which contain this typecode.
corba_tc_encode_with :: TypeCode -> [String] -> IO TypeCodePtr

corba_tc_encode_with (Tk_Null) repoids 		= corba_tc_create_basic tk_null 
corba_tc_encode_with (Tk_Void) repoids		= corba_tc_create_basic tk_void 
corba_tc_encode_with (Tk_Short) repoids		= corba_tc_create_basic tk_short 
corba_tc_encode_with (Tk_Long) repoids		= corba_tc_create_basic tk_long 
corba_tc_encode_with (Tk_UShort) repoids	= corba_tc_create_basic tk_ushort 
corba_tc_encode_with (Tk_ULong) repoids		= corba_tc_create_basic tk_ulong 
corba_tc_encode_with (Tk_Float) repoids		= corba_tc_create_basic tk_float 
corba_tc_encode_with (Tk_Double) repoids	= corba_tc_create_basic tk_double 
corba_tc_encode_with (Tk_Boolean) repoids	= corba_tc_create_basic tk_boolean 
corba_tc_encode_with (Tk_Char) repoids		= corba_tc_create_basic tk_char 
corba_tc_encode_with (Tk_Octet) repoids		= corba_tc_create_basic tk_octet 
corba_tc_encode_with (Tk_Any) repoids		= corba_tc_create_basic tk_any 
corba_tc_encode_with (Tk_String bound) repoids 	= corba_tc_create_string bound 
corba_tc_encode_with (Tk_TypeCode) repoids	= corba_tc_create_basic tk_typecode
corba_tc_encode_with (Tk_ObjRef repoid name) repoids = corba_tc_create_objref repoid name 

corba_tc_encode_with (Tk_Struct repoid name members) repoids
	| elem repoid repoids
	= corba_tc_create_recursive repoid 
	| otherwise
	= do
		seq <- corba_struct_member_seq_encode_with members (repoid:repoids)
		corba_tc_create_struct repoid name seq 

corba_tc_encode_with (Tk_Sequence bound member_type) repoids
	= do
		tcPtr <- corba_tc_encode_with member_type repoids
		corba_tc_create_sequence bound tcPtr

corba_tc_encode_with (Tk_Union repoid name label_type members) repoids
	| elem repoid repoids
	= corba_tc_create_recursive repoid 
	| otherwise
	= do
		labelPtr <- corba_tc_encode_with label_type repoids
		seq	<- corba_union_member_seq_encode_with members (repoid:repoids)
		corba_tc_create_union repoid name labelPtr seq 

corba_tc_encode_with (Tk_Enum repoid name members) repoids
	= do
		seq <- corba_enum_member_seq_encode members 
		corba_tc_create_enum repoid name seq 

corba_tc_encode_with (Tk_Alias repoid name content_type) repoids
	= do
		contentTcPtr <- corba_tc_encode_with content_type repoids
		corba_tc_create_alias repoid name contentTcPtr

corba_tc_encode_with (Tk_Array length member_type) repoids
	= do
		memberTcPtr <- corba_tc_encode_with member_type repoids
		corba_tc_create_array memberTcPtr length 

corba_tc_encode_with (Tk_Exception repoid name members) repoids
	= do
		seq <- corba_struct_member_seq_encode_with members repoids
		corba_tc_create_exception repoid name seq 

corba_tc_encode_with (Tk_Recursive repoid) repoids
	= corba_tc_create_recursive repoid 


corba_tc_decode :: TypeCodePtr -> IO TypeCode
corba_tc_decode tc 
	= do
		kind <- corba_tc_get_kind tc 
		decode_kind tc kind 
	where
	decode_kind :: TypeCodePtr -> Int -> IO TypeCode
	decode_kind tc kind 
		| kind == tk_null  		= return Tk_Null
		| kind == tk_void 	 	= return Tk_Void
		| kind == tk_short  	= return Tk_Short
		| kind == tk_long 	 	= return Tk_Long
		| kind == tk_ushort  	= return Tk_UShort
		| kind == tk_ulong  	= return Tk_ULong
		| kind == tk_float  	= return Tk_Float
		| kind == tk_double  	= return Tk_Double
		| kind == tk_boolean  	= return Tk_Boolean
		| kind == tk_char 	 	= return Tk_Char
		| kind == tk_octet  	= return Tk_Octet
		| kind == tk_any	 	= return Tk_Any

		| kind == tk_string 
		= do
			length <- corba_tc_get_length tc
			return (Tk_String length)

		| kind == tk_typecode 	= return Tk_TypeCode

		| kind == tk_objref 
		= do
			repoid 	<- corba_tc_get_id tc 
			name 	<- corba_tc_get_name tc 
			return (Tk_ObjRef repoid name)

		| kind == tk_struct  = decode_struct tc 

	--
	-- This function never terminates if the sequence is recursive.
	-- To avoid this, we check if the content type is a struct or union,
	-- and return a Tk_Recursive type code instead of decoding the content
	-- type.
	--

		| kind == tk_sequence 
		= do
			length			<- corba_tc_get_length tc 
			contentTypePtr 	<- corba_tc_get_content_type tc 
			content_kind	<- corba_tc_get_kind contentTypePtr 
			if (content_kind == tk_struct || content_kind == tk_union) 
				then do
					content_repoid 	<- corba_tc_get_id contentTypePtr 
					return (Tk_Sequence length (Tk_Recursive content_repoid))
				else do
					content_type	<- corba_tc_decode contentTypePtr 
					return (Tk_Sequence length content_type)

		| kind == tk_union  = decode_union tc 

		| kind == tk_enum  = decode_enum tc 

		| kind == tk_alias 
		= do
			repoid	<- corba_tc_get_id tc 
			name	<- corba_tc_get_name tc 
			contentTypePtr 	<- corba_tc_get_content_type tc 
			content_type	<- corba_tc_decode contentTypePtr 
			return (Tk_Alias repoid name content_type)

		| kind == tk_array 
		= do
			contentTypePtr 	<- corba_tc_get_content_type tc 
			length			<- corba_tc_get_length tc 
			content_type	<- corba_tc_decode contentTypePtr 
			return (Tk_Array length content_type)

		| kind == tk_except  = decode_exception tc 

		| otherwise	    	 = error "unknown TCKind in decode_kind."

	-- TypeCodes for structs are created using StructMemberSeq-s, and
	-- queried using functions of the TypeCode object... :(
	decode_struct :: TypeCodePtr -> IO TypeCode
	decode_struct tc 
		= do
			repoid	<- corba_tc_get_id tc 
			name	<- corba_tc_get_name tc 
			member_count <- corba_tc_get_member_count tc 
			members	<- decode_corba_seq get_member tc member_count 
			return (Tk_Struct repoid name members)
		where
		get_member :: TypeCodePtr -> Int -> IO (String, TypeCode)
		get_member tc index 
			= do
				member_name	<- corba_tc_get_member_name tc index 
				memberTypePtr <- corba_tc_get_member_type tc index 
				member_type	<- corba_tc_decode memberTypePtr 
				return (member_name, member_type)

	-- Same as decode_struct
	decode_exception :: TypeCodePtr -> IO TypeCode
	decode_exception tc 
		= do
			repoid	<- corba_tc_get_id tc 
			name	<- corba_tc_get_name tc 
			member_count <- corba_tc_get_member_count tc 
			members	<- decode_corba_seq get_member tc member_count 
			return (Tk_Exception repoid name members)
		where
		get_member :: TypeCodePtr -> Int -> IO (String, TypeCode)
		get_member tc index 
			= do
				member_name	<- corba_tc_get_member_name tc index 
				memberTypePtr <- corba_tc_get_member_type tc index 
				member_type	<- corba_tc_decode memberTypePtr 
				return (member_name, member_type)

	decode_union :: TypeCodePtr -> IO TypeCode
	decode_union tc 
		= do
			repoid			<- corba_tc_get_id tc 
			name			<- corba_tc_get_name tc 
			member_count 	<- corba_tc_get_member_count tc 
			members			<- decode_corba_seq get_member tc member_count 
			discTypePtr		<- corba_tc_get_discriminator_type tc 
			discType		<- corba_tc_decode discTypePtr 
			return (Tk_Union repoid name discType members)
		where
		get_member :: TypeCodePtr -> Int -> IO (String, Any, TypeCode)
		get_member tc index 
			= do
				member_name <- corba_tc_get_member_name tc index 
				labelPtr	<- corba_tc_get_member_label tc index 
				typePtr		<- corba_tc_get_member_type tc index 
				member_label <- corba_any_decode labelPtr 
				member_type	<- corba_tc_decode typePtr 
				return (member_name, member_label, member_type)

	decode_enum :: TypeCodePtr -> IO TypeCode
	decode_enum tc 
		= do
			repoid	<- corba_tc_get_id tc 
			name	<- corba_tc_get_name tc 
			member_count <- corba_tc_get_member_count tc 
			members	<- decode_corba_seq get_member tc member_count 
			return (Tk_Enum repoid name members)
		where
		get_member :: TypeCodePtr -> Int -> IO String
		get_member tc index = corba_tc_get_member_name tc index 

--------------------------------
--          CORBA::Any
--------------------------------

-- HIGH LEVEL INTERFACE

data Any = 	CorbaNull | 
			CorbaVoid | 
			CorbaShort Int | 
			CorbaLong Int |
		   	CorbaUShort Int | 
			CorbaULong Int | 
			CorbaFloat Double |
			CorbaDouble Double |
			CorbaBoolean Bool |
			CorbaChar Char |
			CorbaOctet Int |
			CorbaString String |
			CorbaAny Any |
			CorbaTypeCode TypeCode |
			CorbaObjRef CORBA_Object |
			CorbaStruct TypeCode [(String, Any)] |  -- tc [(name, type)]
			CorbaSequence TypeCode [Any] |			-- tc [members]
			CorbaUnion TypeCode Any Any	|			-- tc discriminator value
			CorbaEnum TypeCode String |				-- tc value
			CorbaArray TypeCode [Any] |				-- tc [members]
			CorbaException TypeCode [(String, Any)]
	deriving Show

--
-- If there is no active element in a CorbaUnion, the second argument
-- is CorbaNull.
--

corba_nv_pair_seq_encode :: [(String, Any)] -> IO NameValuePairSeq
corba_nv_pair_seq_encode members 
	= encode_corba_seq corba_nv_pair_seq_create add_member members 
	where
		add_member :: NameValuePairSeq -> Int -> (String, Any) -> IO ()
		add_member seq index (name, value) 
			= do
				valuePtr	<- corba_any_encode value 
				corba_nv_pair_seq_fill seq index name valuePtr 

corba_nv_pair_seq_decode :: NameValuePairSeq -> IO [(String, Any)]
corba_nv_pair_seq_decode seq 
	= do
		len	<- corba_nv_pair_seq_length seq 
		decode_corba_seq decode_member seq len 
	where
		decode_member :: NameValuePairSeq -> Int -> IO (String, Any)
		decode_member seq index 
			= do
				(name, valuePtr) 	<- corba_nv_pair_seq_get seq index 
				value				<- corba_any_decode valuePtr 
				return (name, value)

corba_any_seq_encode :: [Any] -> IO AnySeq
corba_any_seq_encode members 
	= encode_corba_seq (corba_any_seq_create) (add_member) members 
	where
		add_member :: AnySeq -> Int -> Any -> IO ()
		add_member seq index value
			= do
				valuePtr	<- corba_any_encode value 
				corba_any_seq_fill seq index valuePtr 
				corba_any_delete valuePtr 

corba_any_seq_decode :: AnySeq -> IO [Any]
corba_any_seq_decode seq 
	= do
		len <- corba_any_seq_length seq 
		decode_corba_seq decode_member seq len 
		where
		decode_member :: AnySeq -> Int -> IO Any
		decode_member seq index 
			= do
				valuePtr	<- corba_any_seq_get seq index 
				value		<- corba_any_decode valuePtr 
				return value

corba_any_encode :: Any -> IO AnyPtr
corba_any_encode any  = create any
	where
	create (CorbaNull) 			= corba_any_create_null 
	create (CorbaVoid) 			= 
				error "corba_any_create called with CorbaVoid argument"
	create (CorbaShort s) 		= corba_any_create_short s 
	create (CorbaUShort s) 		= corba_any_create_ushort s 
	create (CorbaLong s) 		= corba_any_create_long s 
	create (CorbaULong s) 		= corba_any_create_ulong s 
	create (CorbaFloat f) 		= corba_any_create_float f 
	create (CorbaDouble d) 		= corba_any_create_double d 
	create (CorbaBoolean b) 	= corba_any_create_bool b 
	create (CorbaChar c) 		= corba_any_create_char c 
	create (CorbaOctet s)		= corba_any_create_octet s 
	create (CorbaString s)		= corba_any_create_string s 

	create (CorbaAny any)
		= do
			anyPtr	<- corba_any_encode any 
			any 	<- corba_any_create_any anyPtr
			return any

	create (CorbaTypeCode tc)
		= do
			tcPtr	<- corba_tc_encode tc 
			tc		<- corba_any_create_tc tcPtr
			return tc

	create (CorbaObjRef (CORBA_Object ref))
		= corba_any_create_objref ref

	create (CorbaStruct tc members)
		= do
			tcPtr		<- corba_tc_encode tc 
			membersPtr	<- corba_nv_pair_seq_encode members 
			corba_dyn_struct_create tcPtr membersPtr 

	create (CorbaSequence tc elements)
		= do
			tcPtr		<- corba_tc_encode tc 
			elementsPtr	<- corba_any_seq_encode elements 
			corba_dyn_sequence_create tcPtr elementsPtr 

	create (CorbaUnion tc discriminant value)
		= do
			tcPtr		<- corba_tc_encode tc 
			discriminantPtr <- corba_any_encode discriminant 
			valuePtr	<- corba_any_encode value 
			corba_dyn_union_create tcPtr discriminantPtr valuePtr 

	create (CorbaEnum tc value)
		= do
			tcPtr		<- corba_tc_encode tc 
			corba_dyn_enum_create tcPtr value 

	create (CorbaArray tc elements)
		= do
			tcPtr		<- corba_tc_encode tc 
			elementsPtr	<- corba_any_seq_encode elements 
			corba_dyn_array_create tcPtr elementsPtr 

	-- Same as for CorbaStructs...
	create (CorbaException tc members)
		= do
			tcPtr		<- corba_tc_encode tc 
			membersPtr	<- corba_nv_pair_seq_encode members 
			corba_dyn_struct_create tcPtr membersPtr 

corba_any_decode :: AnyPtr -> IO Any
corba_any_decode any 
	= do
		tcPtr	<- corba_any_get_type any 
		tc		<- corba_tc_decode tcPtr 
		decode any tc 
	where
	decode any (Tk_Null)   = return (CorbaNull)
	decode any (Tk_Void)   = return (CorbaVoid)

	decode any (Tk_Short)  
		= do
			value	<- corba_any_extract_short any 
			return (CorbaShort value)

	decode any (Tk_Long)  
		= do
			value	<- corba_any_extract_long any 
			return (CorbaLong value)

	decode any (Tk_UShort)  
		= do
			value	<- corba_any_extract_ushort any 
			return (CorbaUShort value)

	decode any (Tk_ULong) 
		= do
			value	<- corba_any_extract_ulong any 
			return (CorbaULong value)

	decode any (Tk_Float) 
		= do
			value	<- corba_any_extract_float any
			return (CorbaFloat value)

	decode any (Tk_Double)  
		= do
			value	<- corba_any_extract_double any 
			return (CorbaDouble value)

	decode any (Tk_Boolean) 
		= do
			value	<- corba_any_extract_boolean any 
			return (CorbaBoolean value)

	decode any (Tk_Char) 
		= do
			value	<- corba_any_extract_char any 
			return (CorbaChar value)

	decode any (Tk_Octet) 
		= do
			value	<- corba_any_extract_octet any 
			return (CorbaOctet value)

	decode any (Tk_String l) 
		= do
			value	<- corba_any_extract_string any 
			return (CorbaString value)

	decode any (Tk_Any) 
		= do
			anyPtr	<- corba_any_extract_any any 
			value	<- corba_any_decode anyPtr 
			return (CorbaAny value)

	decode any (Tk_TypeCode) 
		= do
			tcPtr	<- corba_any_extract_tc any 
			value	<- corba_tc_decode tcPtr 
			return (CorbaTypeCode value)

	decode any (Tk_ObjRef repoid name)  
		= do
			ref		<- corba_any_extract_objref any 
			return (CorbaObjRef (CORBA_Object ref))

	decode any (Tk_Struct repoid name members) 
		= do
			seqPtr	<- corba_dyn_struct_members any 
			seq		<- corba_nv_pair_seq_decode seqPtr 
			corba_nv_pair_seq_delete seqPtr 
			return (CorbaStruct (Tk_Struct repoid name members) seq)

	decode any (Tk_Exception repoid name members) 
		= do
			seqPtr	<- corba_dyn_struct_members any 
			seq		<- corba_nv_pair_seq_decode seqPtr 
			corba_nv_pair_seq_delete seqPtr 
			return (CorbaException (Tk_Exception repoid name members) seq)

	decode any (Tk_Sequence bound tc) 
		= do
			seqPtr	<- corba_dyn_sequence_elements any 
			seq		<- corba_any_seq_decode seqPtr 
			corba_any_seq_delete seqPtr 
			return (CorbaSequence (Tk_Sequence bound tc) seq)

	decode any (Tk_Union repoid name tc members) 
		= do
			(labelPtr, valuePtr) 	<- corba_dyn_union_get any 
			label					<- corba_any_decode labelPtr 
			value					<- corba_any_decode valuePtr 
			return (CorbaUnion (Tk_Union repoid name tc members) label value)

	decode any (Tk_Enum repoid name members) 
		= do
			value	<- corba_dyn_enum_get any 
			return (CorbaEnum (Tk_Enum repoid name members) value)

	decode any (Tk_Array bound tc) 
		= do
			seqPtr	<- corba_dyn_array_elements any 
			seq		<- corba_any_seq_decode seqPtr 
			corba_any_seq_delete seqPtr 
			return (CorbaArray (Tk_Array bound tc) seq)

	decode any (Tk_Alias _ _ tc)
		= decode any tc

	decode any tc  = error ("corba_any_decode called with " ++ (show tc) ++ " argument.")
	
--------------------------------
--        EXCEPTION
--------------------------------

-- HIGHER LEVEL INTERFACE

data CompletionStatus = Completed_Yes | Completed_No | Completed_Maybe
	 deriving (Show);
data CorbaException = NoException | 
					SystemException String CompletionStatus Int | 
					UserException String Any
	 deriving (Show);

decode_corba_exception :: ExceptionPtr -> [TypeCode] -> IO CorbaException
decode_corba_exception ex exception_defs 
	= do
		is_system	<- corba_exception_is_system ex 
		repoid	<- corba_exception_get_repoid ex 
		if is_system 
			then do
			minor	<- corba_exception_get_minor ex 
			completed <- corba_exception_get_completed ex 
			return (SystemException repoid (decode_completed completed) minor)
			else
			decode_user_exception ex repoid 
		where
		decode_user_exception ex repoid 
			= do
				unknown	<- corba_unknown_ex_narrow ex 
				maybe f2 (\e -> f1 e) unknown
			where
				f1 unknown = do
					repoid	<- corba_unknown_ex_get_repoid unknown
					decode_unknown unknown repoid 
				f2 = return (UserException repoid (CorbaNull))
					
				-- Find the exception definition corresponding to 'repoid',
				-- and try to decode the contents of the exception according to
				-- it.
				decode_unknown unknown repoid 
					= do
						if null defs 
							then do
							return (UserException repoid (CorbaNull))
							else do
							tcPtr	<- corba_tc_encode (head defs) 
							valuePtr <- corba_unknown_ex_get_value unknown tcPtr 
							value	<- corba_any_decode valuePtr 
							return (UserException repoid value)
						where
							defs = filter find_def exception_defs
							find_def (Tk_Exception repoid2 _ _) = (repoid == repoid2)
							find_def any = error ("find_def called with non exception argument: " ++ (show any))

		decode_completed :: Int -> CompletionStatus
		decode_completed i
			| i == completed_yes 	= Completed_Yes
			| i == completed_no 	= Completed_No
			| i == completed_maybe 	= Completed_Maybe
			| otherwise				= error "bad value for CompletionStatus."

corba_exception_encode :: CorbaException -> Any
corba_exception_encode (UserException repoid ex) = ex

--------------------------------
--         NAMEDVALUE
--------------------------------

data CorbaArg = InArg String Any 
			| 	OutArg String Any 
			| 	InOutArg String Any
			| 	InArgTemplate String TypeCode
			| 	OutArgTemplate String TypeCode
			| 	InOutArgTemplate String TypeCode
	deriving Show

corba_arg_get_value :: CorbaArg -> Any
corba_arg_get_value (InArg _ v) 	= v
corba_arg_get_value (OutArg _ v) 	= v
corba_arg_get_value (InOutArg _ v) 	= v
corba_arg_get_value arg				= error ("corba_arg_get_value called with " ++ (show arg) ++ " argument.")

corba_nv_decode :: NamedValuePtr -> IO CorbaArg
corba_nv_decode nv 
	= do
		name	<- corba_nv_get_name nv 
		valuePtr <- corba_nv_get_value nv 
		flags	<- corba_nv_get_flags nv 
		value	<- corba_any_decode valuePtr 
		return (decode name value flags)
	where
		decode :: String -> Any -> Int -> CorbaArg
		decode name value argtype
			| argtype == arg_in 	= InArg name value
			| argtype == arg_out 	= OutArg name value
			| argtype == arg_inout 	= InOutArg name value
			| otherwise			 	= error "bad flags."

corba_nvlist_decode :: NVListPtr -> IO [CorbaArg]
corba_nvlist_decode nvlist 
	= do
		len	<- corba_nvlist_count nvlist 
		decode_corba_seq decode_member nvlist len 
		where
			decode_member nvlist index 
			  = do
				nvPtr	<- corba_nvlist_get nvlist index 
				corba_nv_decode nvPtr 

corba_nvlist_encode :: [CorbaArg] -> IO NVListPtr
corba_nvlist_encode members 
	= encode_corba_seq (\i -> corba_nvlist_create) (add_member) members 
	where
		add_member :: NVListPtr -> Int -> CorbaArg -> IO ()

		add_member seq index (InArg name value) 
			= do
				nv	<- corba_nvlist_add seq arg_in 
				valuePtr <- corba_any_encode value 
				corba_nv_set_value nv valuePtr 

		add_member seq index (OutArg name value) 
			= do
				nv	<- corba_nvlist_add seq arg_out
				valuePtr <- corba_any_encode value 
				corba_nv_set_value nv valuePtr 

		add_member seq index (InOutArg name value) 
			= do
				nv	<- corba_nvlist_add seq arg_inout
				valuePtr <- corba_any_encode value 
				corba_nv_set_value nv valuePtr 

		add_member seq index (InArgTemplate name tc) 
			= do
				nv		<- corba_nvlist_add seq arg_in 
				tcPtr 	<- corba_tc_encode tc 
				corba_nv_set_type nv tcPtr 

		add_member seq index (OutArgTemplate name tc)
			= do
				nv		<- corba_nvlist_add seq arg_out
				tcPtr 	<- corba_tc_encode tc 
				corba_nv_set_type nv tcPtr 

		add_member seq index (InOutArgTemplate name tc) 
			= do
				nv		<- corba_nvlist_add seq arg_inout
				tcPtr 	<- corba_tc_encode tc 
				corba_nv_set_type nv tcPtr 

--------------------------------
--          REQUEST
--------------------------------

corba_invoke :: CORBA_Object -> String -> [CorbaArg] -> TypeCode -> [TypeCode] -> IO (Any, CorbaException, [CorbaArg])
corba_invoke (CORBA_Object obj) operation args result_tc raises 
	= do
		req	<- corba_object_create_request obj operation 
		resultTcPtr <- corba_tc_encode result_tc 
		corba_request_set_result_tc req resultTcPtr 
	  	sequence_ (map (add_arg req)  args)
	  	sequence_ (map (add_exception req)  raises)
	  	corba_request_invoke req 
		ex	<- corba_request_exception req 
		if is_just ex
			then do
			ex <- decode_corba_exception (just ex) raises
			return (CorbaVoid, ex, [])
			else do
			resPtr <- corba_request_result req 
			params <- get_out_params req 
			res <- corba_any_decode resPtr 
	 		corba_request_delete req 
			return (res, NoException, params)
	where
		-- Add arguments to the request
		add_arg :: RequestPtr -> CorbaArg -> IO ()
		add_arg req (InArg name value) 
			= do
				valuePtr <- corba_any_encode value 		
		 	  	corba_request_add_in_arg req name valuePtr 
			  	corba_any_delete valuePtr 

		add_arg req (OutArg name value) 
			= do
				valuePtr <- corba_any_encode value 		
		 	  	corba_request_add_out_arg req name valuePtr 
		  		corba_any_delete valuePtr 

		add_arg req (InOutArg name value) 
			= do
				valuePtr <- corba_any_encode value 		
		 	  	corba_request_add_inout_arg req name valuePtr 
		  		corba_any_delete valuePtr 

		add_arg req (OutArgTemplate name tc) 
			= do	
				tcPtr	<- corba_tc_encode tc 
				corba_request_add_out_arg_template req name tcPtr 

			-- Add a possible exception
		add_exception :: RequestPtr -> TypeCode -> IO ()
		add_exception req tc 
			= do
				tcPtr	<- corba_tc_encode tc 
			 	corba_request_add_exception_def req tcPtr 

		-- Collect the out and inout arguments
		get_out_params req 
			= do
				nvlistPtr	<- corba_request_arguments req 
				nvlist		<- corba_nvlist_decode nvlistPtr 
				return (filter is_out_param nvlist)

		is_out_param :: CorbaArg -> Bool
		is_out_param (InArg _ _) = False
		is_out_param _			= True


----------------------------------------------------------------------
--             DEFINITIONS USED BY THE GENERATED CODE
----------------------------------------------------------------------

--------------------------------
--        CORBA::Void
--------------------------------

type CORBA_Void	=	()

corba_Void__tc	::	TypeCode
corba_Void__tc	=	Tk_Void

corba_Void__decode :: Any -> IO (Maybe CORBA_Void)
corba_Void__decode (CorbaVoid)	 =	return (Just ())
corba_Void__decode _ 			=	return (Nothing)

corba_Void__encode :: CORBA_Void -> Any
corba_Void__encode _ = (CorbaVoid)

--------------------------------
--        CORBA::Short
--------------------------------

type CORBA_Short	=	Int

corba_Short__tc	::	TypeCode
corba_Short__tc	=	Tk_Short

corba_Short__decode :: Any -> IO (Maybe CORBA_Short)
corba_Short__decode (CorbaShort i)		=	return (Just i)
corba_Short__decode _ 					=	return (Nothing)

corba_Short__encode :: CORBA_Short -> Any
corba_Short__encode i = (CorbaShort i)

--------------------------------
--        CORBA::Long
--------------------------------

type CORBA_Long	=	Int

corba_Long__tc	::	TypeCode
corba_Long__tc	=	Tk_Long

corba_Long__decode :: Any -> IO (Maybe CORBA_Long)
corba_Long__decode (CorbaLong i) 	=	return (Just i)
corba_Long__decode _ 				=	return (Nothing)

corba_Long__encode :: CORBA_Long -> Any
corba_Long__encode i = (CorbaLong i)

--------------------------------
--        CORBA::UShort
--------------------------------

type CORBA_UShort	=	Int

corba_UShort__tc	::	TypeCode
corba_UShort__tc	=	Tk_UShort

corba_UShort__decode :: Any -> IO (Maybe CORBA_UShort)
corba_UShort__decode (CorbaUShort i) 		= return (Just i)
corba_UShort__decode _						= return (Nothing)

corba_UShort__encode :: CORBA_UShort -> Any
corba_UShort__encode i = (CorbaUShort i)

--------------------------------
--        CORBA::ULong
--------------------------------

type CORBA_ULong	=	Int

corba_ULong__tc	::	TypeCode
corba_ULong__tc	=	Tk_ULong

corba_ULong__decode :: Any -> IO (Maybe CORBA_ULong)
corba_ULong__decode (CorbaULong i) 	= return (Just i)
corba_ULong__decode _ 				= return (Nothing)

corba_ULong__encode :: CORBA_ULong -> Any
corba_ULong__encode i = (CorbaULong i)

--------------------------------
--        CORBA::Float
--------------------------------

type CORBA_Float	=	Double

corba_Float__tc	::	TypeCode
corba_Float__tc	=	Tk_Float

corba_Float__decode :: Any 	->	IO (Maybe CORBA_Float)
corba_Float__decode (CorbaFloat f) 	= return (Just f)
corba_Float__decode _ 				= return (Nothing)

corba_Float__encode :: CORBA_Float -> Any
corba_Float__encode f = (CorbaFloat f)

--------------------------------
--        CORBA::Double
--------------------------------

type CORBA_Double	=	Double

corba_Double__tc	::	TypeCode
corba_Double__tc	=	Tk_Double

corba_Double__decode :: Any 	->	IO (Maybe CORBA_Double)
corba_Double__decode (CorbaDouble f) 	= return (Just f)
corba_Double__decode _ 					= return (Nothing)

corba_Double__encode :: CORBA_Double -> Any
corba_Double__encode f = (CorbaDouble f)

--------------------------------
--        CORBA::Boolean
--------------------------------

type CORBA_Boolean	=	Bool

corba_Boolean__tc	::	TypeCode
corba_Boolean__tc	=	Tk_Boolean

corba_Boolean__decode :: Any 	->	IO (Maybe CORBA_Boolean)
corba_Boolean__decode (CorbaBoolean b) 	= return (Just b)
corba_Boolean__decode _ 				= return (Nothing)

corba_Boolean__encode :: CORBA_Boolean -> Any
corba_Boolean__encode b = (CorbaBoolean b)

--------------------------------
--        CORBA::Char
--------------------------------

type CORBA_Char	=	Char

corba_Char__tc	::	TypeCode
corba_Char__tc	=	Tk_Char

corba_Char__decode :: Any 	->	IO (Maybe CORBA_Char)
corba_Char__decode	(CorbaChar c) 	=	return (Just c)
corba_Char__decode	_ 				=	return (Nothing)

corba_Char__encode :: CORBA_Char -> Any
corba_Char__encode c = (CorbaChar c)

--------------------------------
--        CORBA::Octet
--------------------------------

type CORBA_Octet	=	Int

corba_Octet__tc	::	TypeCode
corba_Octet__tc	=	Tk_Octet

corba_Octet__decode :: Any 	->	IO (Maybe CORBA_Octet)
corba_Octet__decode (CorbaOctet i) 	=	return (Just i)
corba_Octet__decode _ 				=	return (Nothing)

corba_Octet__encode :: CORBA_Octet -> Any
corba_Octet__encode i = (CorbaOctet i)

--------------------------------
--        CORBA::Any
--------------------------------

type CORBA_Any	=	Any

corba_Any__tc	::	TypeCode
corba_Any__tc	=	Tk_Any

corba_Any__decode :: Any 	->	IO (Maybe CORBA_Any)
corba_Any__decode (CorbaAny any) 	=	return (Just any)
corba_Any__decode _ 				=	return (Nothing)

corba_Any__encode :: Any -> Any
corba_Any__encode any = (CorbaAny any)

--------------------------------
--     CORBA::TypeCode
--------------------------------

type CORBA_TypeCode	=	TypeCode

corba_TypeCode__tc :: TypeCode
corba_TypeCode__tc	=	Tk_TypeCode

corba_TypeCode__decode :: Any -> IO (Maybe CORBA_TypeCode)
corba_TypeCode__decode (CorbaTypeCode tc) 	= return (Just tc)
corba_TypeCode__decode _ 					= return (Nothing)

corba_TypeCode__encode :: TypeCode -> Any
corba_TypeCode__encode tc = (CorbaTypeCode tc)

--------------------------------
--      CORBA::String
--------------------------------

-- This is the unbounded version
type CORBA_String	=	String

corba_String__tc :: TypeCode
corba_String__tc = Tk_String 0

corba_String__decode :: Any -> IO (Maybe CORBA_String)
corba_String__decode (CorbaString s)  	= return (Just s)
corba_String__decode _ 					= return (Nothing)

corba_String__encode :: CORBA_String -> Any
corba_String__encode s = (CorbaString s)

--------------------------------
--      CORBA::Object
--------------------------------

data CORBA_Object  = CORBA_Object CorbaObjectPtr
	deriving Show

corba_Object__create :: CORBA_Object -> CORBA_Object
corba_Object__create obj = obj

corba_Object__tc :: TypeCode
corba_Object__tc = (Tk_ObjRef "IDL:omg.org/CORBA/Object:1.0" "Object")

corba_Object__decode :: Any -> IO (Maybe CORBA_Object)
corba_Object__decode (CorbaObjRef ref)  = return (Just ref)
corba_Object__decode _					= return (Nothing)

corba_Object__encode :: CORBA_Object -> Any
corba_Object__encode obj = (CorbaObjRef obj)

corba_Object_is_a :: CORBA_Object -> String -> IO Bool
corba_Object_is_a (CORBA_Object ref) repoid  = corba_object_is_a ref repoid 

corba_Object__nil :: IO CORBA_Object
corba_Object__nil 
	= do
		ref <- corba_object_nil
		return (CORBA_Object ref)

class CORBA_Object__widen_class a where
	corba_Object__widen :: a -> CORBA_Object

instance CORBA_Object__widen_class CORBA_Object where
	corba_Object__widen obj = obj

--------------------------------
--         ORB
--------------------------------

type CORBA_ORB = ORBPtr

corba_ORB_init :: [String] -> IO (CORBA_ORB, [String])
corba_ORB_init args 
	= do
		-- MICO needs a dummy first argument (executable name?)
		(argv, argc)	<- c_string_array_encode args
		(orb, out_argv, out_argc) <- corba_orb_init (argv, argc) 
		out_args		<- c_string_array_decode out_argv out_argc 
		return (orb, out_args)

corba_ORB_bind :: CORBA_ORB -> String -> String -> IO (Maybe CORBA_Object)
corba_ORB_bind orb address repoid 
	= do
		obj	<- corba_orb_bind orb address repoid
		return (maybe Nothing (\ref -> Just (CORBA_Object ref)) obj)

corba_ORB_string_to_object :: CORBA_ORB -> String -> IO (Maybe CORBA_Object)
corba_ORB_string_to_object orb ior 
	= do
		obj	<- corba_orb_string_to_object orb ior 
		return (maybe Nothing (\ref -> Just (CORBA_Object ref)) obj)

corba_ORB_object_to_string :: CORBA_ORB -> CORBA_Object -> IO String
corba_ORB_object_to_string orb (CORBA_Object ref)
	= corba_orb_object_to_string orb ref

corba_ORB_resolve_initial_references :: CORBA_ORB -> String -> IO (Maybe CORBA_Object)
corba_ORB_resolve_initial_references orb name 
	= do
		obj	<- corba_orb_resolve_initial_references orb name 
		return (maybe Nothing (\ref -> Just (CORBA_Object ref)) obj)

writeIORToFile :: CORBA_ORB -> CORBA_Object -> String -> IO ()
writeIORToFile orb obj filename = do
	file <- openFile filename WriteMode
	ior <- corba_ORB_object_to_string orb obj
	hPutStrLn file ior
	hClose file

readIORFromFile :: CORBA_ORB -> String -> IO (Maybe CORBA_Object)
readIORFromFile orb filename = do
	file <- openFile filename ReadMode
	ior <-  hGetLine file
	hClose file
	corba_ORB_string_to_object orb ior

--------------------------------
--      SERVER REQUEST
--------------------------------

type CORBA_ServerRequest = ServerRequestPtr

corba_ServerRequest_operation :: CORBA_ServerRequest -> IO String
corba_ServerRequest_operation req = corba_server_request_operation req

corba_ServerRequest_set_arg_types :: CORBA_ServerRequest -> [CorbaArg] -> IO ()
corba_ServerRequest_set_arg_types req args
	= do
	  nvlist	<- corba_nvlist_encode args
	  corba_server_request_set_params req nvlist

corba_ServerRequest_get_arg :: CORBA_ServerRequest -> Int -> IO Any
corba_ServerRequest_get_arg req index
	= do
		nvlist 	<- corba_server_request_get_args req
		nv		<- corba_nvlist_get nvlist index
		anyPtr 	<- corba_nv_get_value nv
		any		<- corba_any_decode anyPtr
		return any

corba_ServerRequest_set_arg :: CORBA_ServerRequest -> Int -> Any -> IO ()
corba_ServerRequest_set_arg req index value
	= do
		nvlist 	<- corba_server_request_get_args req
		nv		<- corba_nvlist_get nvlist index
		anyPtr 	<- corba_any_encode value
	 	corba_nv_set_value nv anyPtr

corba_ServerRequest_set_result :: CORBA_ServerRequest -> Any -> IO ()
corba_ServerRequest_set_result req result
	= do
		anyPtr <- corba_any_encode result
		corba_server_request_set_result req anyPtr

corba_ServerRequest_set_exception :: CORBA_ServerRequest -> Any -> IO ()
corba_ServerRequest_set_exception req ex
	= do
		anyPtr <- corba_any_encode ex
		corba_server_request_set_exception req anyPtr

--------------------------------
--        SERVER SIDE
--------------------------------

data PSt l = PSt
	{	ls :: l
	,	io :: IOSt l
	}


data IOSt l = IOSt {
	servants 		:: [(CorbaObjectPtr, CORBA_Servant_Wrapper l)],
	need_to_exit 	:: Bool,
	current			:: CORBA_Object,	-- The currently executing object
	orb				:: CORBA_ORB
	}

type ServerInitFunc l = (PSt l)  -> IO (PSt l)

noInit :: (PSt l) -> IO (PSt l)
noInit ps  = return ps

type ServantDispatchFun l = (PSt l) -> CORBA_ServerRequest -> IO (PSt l)

data CORBA_Servant_Wrapper l = CORBA_Servant_Wrapper (ServantDispatchFun l)

-- I was to too lazy to find out how to do this in GreenCard...
foreign export dynamic mk_corba_server_callback :: (StablePtr (IORef (PSt l)) -> ServerRequestPtr -> CorbaObjectPtr -> IO ()) -> IO Addr

-- corba_server_run callback_ptr ps_ptr
foreign import corba_server_run :: Addr -> (StablePtr (IORef (PSt l))) -> IO ()

corba_Server_run :: CORBA_ORB -> l -> (ServerInitFunc l) -> IO ()
corba_Server_run orb initial_state init_func =
	do
	nil <- corba_object_nil
	ps <- init_func (initial_ps nil)
	cb_addr <- mk_corba_server_callback corba_server_callback
	ps_ref <- newIORef ps
	ps_ref_ptr <- newStablePtr ps_ref
	ps <- corba_server_run cb_addr ps_ref_ptr
	return ()
	where
	initial_ps nil = PSt {
		ls = initial_state,
		io = IOSt {
			servants 		= [],
			need_to_exit	= False,
			current			= CORBA_Object nil,
			orb				= orb
		}
	}

	corba_server_callback :: StablePtr (IORef (PSt l)) -> ServerRequestPtr -> CorbaObjectPtr -> IO ()
	corba_server_callback ps_ptr req obj = do
		ps_ref <- deRefStablePtr ps_ptr
		ps@(PSt _ io@(IOSt servants _ _ _)) <- readIORef ps_ref
		(serv, index) 	<- return $ lookup_servant servants obj 0
		ps				<- return $ ps {io = io { current = CORBA_Object obj }}
		ps				<- dispatch_to ps req serv
		writeIORef ps_ref ps
		if (ps_need_to_exit ps) then corba_server_exit else return ()
		where
		ps_need_to_exit :: (PSt l) -> Bool
		ps_need_to_exit (PSt _ (IOSt _ need_to_exit _ _)) = need_to_exit

		dispatch_to :: (PSt l) -> ServerRequestPtr -> (CORBA_Servant_Wrapper l ) -> IO (PSt l)
		dispatch_to ps req serv@(CORBA_Servant_Wrapper dispatch) = do
			ps <- dispatch ps req
			return ps

		-- The ORB takes care of raising errors for nonexistent servants
		lookup_servant ((o, s):xs) obj index
			| o == obj
			= (s, index)
			| otherwise
			= lookup_servant xs obj (index + 1)
		lookup_servant [] obj index = error "Nonexistent servant."

		-- Update the given elem in the list and return the updated list
		updateAt :: Int -> a -> [a] -> [a]
		updateAt 0 elem (x:xs) = (elem:xs)
		updateAt index elem (x:xs) = (x:(updateAt (index - 1) elem xs))


-- Instruct the CORBA server process to exit
corba_Server_exit :: (PSt l) -> (PSt l)
corba_Server_exit ps@(PSt _ io) = ps {io = io {need_to_exit = True}}

-- Registers a new CORBA servant
corba_Server_add_servant :: (PSt l) -> String -> (CORBA_Servant_Wrapper l ) -> IO (CORBA_Object, (PSt l))
corba_Server_add_servant ps@(PSt _ io) repoid servant
	= do
		obj	<- corba_servant_create repoid
		return ((CORBA_Object obj), ps {io = io {servants = ((obj, servant):servants io)}})

corba_Server_get_current :: (PSt l) -> CORBA_Object
corba_Server_get_current (PSt _ io) = current io

corba_Server_get_orb :: (PSt l) -> CORBA_ORB
corba_Server_get_orb (PSt _ io) = orb io

-----------------------------------------------------
--   UTILITY FUNCTIONS USED BY THE GENERATED CODE
-----------------------------------------------------

-- TODO: Test equivalence of typecodes

corba_enum_decode :: Any -> TypeCode -> (String -> Maybe a) -> IO (Maybe a)
corba_enum_decode (CorbaEnum tc value) enum_tc decode_fun
	| tc /= enum_tc
	= return Nothing
	| otherwise
	= return $ decode_fun value

corba_enum_decode _ _ _ = return Nothing

corba_objref_decode :: Any -> TypeCode -> (CORBA_Object -> a) -> (IO (Maybe a))
corba_objref_decode (CorbaObjRef ref) tc create_fun = return $ Just (create_fun ref)
corba_objref_decode _ _ _ = return Nothing

corba_objref_narrow :: CORBA_Object__widen_class a => a -> String -> (CORBA_Object -> b) -> IO (Maybe b)
corba_objref_narrow obj repoid create_fun = do
	is_a <- corba_Object_is_a (corba_Object__widen obj) repoid
	return (if is_a then (Just (create_fun (corba_Object__widen obj))) else Nothing)

corba_struct_decode :: Any -> TypeCode -> ([(String, Any)] -> IO (Maybe a)) -> IO (Maybe a)
corba_struct_decode (CorbaStruct tc members) struct_tc decode_fun
	| tc /= struct_tc
	= return Nothing
	| otherwise
	= decode_fun members
corba_struct_decode _ _ _ = return Nothing

corba_exception_decode :: Any -> TypeCode -> ([(String, Any)] -> IO (Maybe a)) -> IO (Maybe a)
corba_exception_decode (CorbaException tc members) struct_tc decode_fun
	| tc /= struct_tc
	= return Nothing
	| otherwise
	= decode_fun members
corba_exception_decode _ _ _ = return Nothing

corba_seq_decode :: Any -> TypeCode -> (Any -> IO (Maybe a)) -> IO (Maybe [a])
corba_seq_decode (CorbaSequence tc members) tc2 decode_fun
		| tc /= tc2
		  = return Nothing
		| otherwise
		  = do
			xs <- sequence (map decode_fun members)
			return $ if (and (map is_just xs)) then (Just (map just xs)) else Nothing
corba_seq_decode _ _ _ = return Nothing

corba_array_decode :: Any -> TypeCode -> (Any -> IO (Maybe a)) -> IO (Maybe [a])
corba_array_decode (CorbaArray tc members) tc2 decode_fun
		| tc /= tc2
		  = return Nothing
		| otherwise
		  = do
			xs <- sequence (map decode_fun members)
			return $ if (and (map is_just xs)) then (Just (map just xs)) else Nothing
corba_array_decode _ _ _ = return Nothing

corba_union_decode :: Any -> TypeCode -> (Any -> Any -> IO (Maybe a)) -> IO (Maybe a)
corba_union_decode (CorbaUnion tc discriminator value) tc2 decode_fun
		| tc /= tc2
		  = return Nothing
		| otherwise
		  = decode_fun discriminator value
corba_union_decode _ _ _ = return Nothing

corba_get_objref :: CORBA_Object__widen_class a => a -> CorbaObjectPtr
corba_get_objref obj = ref
	where
	(CORBA_Object ref) = corba_Object__widen obj

narrow_objref :: CORBA_Object__widen_class a => a -> String -> (CORBA_Object -> b) -> IO (Maybe b)
narrow_objref obj repoid narrow_fun = do
				is_a <- corba_Object_is_a obj2 repoid
				return $ if is_a then (Just (narrow_fun obj2)) else (Nothing)
				where
					obj2 = corba_Object__widen obj

corba_invoke_simple :: CORBA_Object -> String -> [CorbaArg] -> TypeCode -> (Any -> IO (Maybe a)) -> IO (ResultOrException a CorbaException)
corba_invoke_simple ref op_name params return_type decode_fun 
	= do
	(result, exception, _) <- corba_invoke ref op_name params return_type []
	result <- decode_fun result
	return (corba_result_create (just result) exception)






-- Local Variables:
-- tab-width: 4
-- End:



