module Main where

import Repository
import Corba
import IO
import IOExts
import List
import Pretty
import System
import Char
import GetOpt

--
--   HIGHER LEVEL INTERFACE TO THE INTERFACE REPOSITORY
--

data IRObject = IRRepository [IRObject]
			| IRModule C_CORBA_ModuleDescription [IRObject]
			| IRConstant C_CORBA_ConstantDescription
			| IRStruct C_CORBA_TypeDescription [IRObject]
			| IRException C_CORBA_ExceptionDescription
			| IRUnion C_CORBA_TypeDescription [IRObject]
			| IREnum C_CORBA_TypeDescription
			| IRAlias C_CORBA_TypeDescription
			| IRInterface C_CORBA_InterfaceDescription [IRObject]
			| IROperation C_CORBA_OperationDescription
			| IRAttribute C_CORBA_AttributeDescription
	deriving Show

ir_dump :: C_CORBA_IRObject -> IO (IRObject)
ir_dump obj = do
	(Result def_kind) <- c_CORBA_IRObject__get_def_kind obj
	dump_by_type obj def_kind
		where
		dump_by_type :: C_CORBA_IRObject -> C_CORBA_DefinitionKind -> IO (IRObject)

		dump_by_type obj Dk_Interface = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_InterfaceDescription__decode desc
			(members) <-get_contents obj 
			return (IRInterface desc members)

		dump_by_type obj Dk_Module = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_ModuleDescription__decode desc 
			(members) <-get_contents obj 
			return (IRModule desc members)

		dump_by_type obj Dk_Repository = do
			(members) <-get_contents obj 
			return (IRRepository members)
	
		dump_by_type obj Dk_Struct = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_TypeDescription__decode desc 
			(members) <-get_contents obj 
			return (IRStruct desc members)

		dump_by_type obj Dk_Union = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_TypeDescription__decode desc 
			(members) <-get_contents obj 
			return (IRUnion desc members)

		dump_by_type obj Dk_Enum = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_TypeDescription__decode desc 
			return (IREnum desc)

		dump_by_type obj Dk_Alias = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_TypeDescription__decode desc 
			return (IRAlias desc)

		dump_by_type obj Dk_Exception = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_ExceptionDescription__decode desc 
			return (IRException desc)

		dump_by_type obj Dk_Operation = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_OperationDescription__decode desc 
			return (IROperation desc)

		dump_by_type obj Dk_Attribute = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_AttributeDescription__decode desc 
			return (IRAttribute desc)

		dump_by_type obj Dk_Constant = do
			(desc) <-get_description obj 
			(Just desc) <-c_CORBA_ConstantDescription__decode desc 
			return (IRConstant desc)

		dump_by_type obj kind = no_rule_match "dump_by_type" kind 

		get_description :: C_CORBA_IRObject -> IO (Any)
		get_description obj = do
			(obj) <-c_CORBA_Contained__narrow obj 
			(Result (C_CORBA_Contained_Description kind value)) <-c_CORBA_Contained_describe (just obj) 
			return value

		get_contents :: C_CORBA_IRObject -> IO ([IRObject])
		get_contents obj = do
			(obj) <-c_CORBA_Container__narrow obj 
			(Result seq) <-c_CORBA_Container_contents (just (obj)) Dk_all True
			(res) <-sequence (map dump_member seq)
			return (res)
				where
			dump_member :: C_CORBA_Contained -> IO (IRObject)
			dump_member obj = do
				(member) <-ir_dump (c_CORBA_IRObject__widen obj) 
				return (member)


--------------------------------
--     UTILITY FUNCTIONS
--------------------------------

no_rule_match :: Show a => String -> a -> b
no_rule_match fct_name value = error ("\n\nError: no pattern matches " ++ (show value) ++ " for function " ++ fct_name ++ "\n\n")

with_indexes :: [a] -> [(a, Int)]
with_indexes l = zip l [0..]

-- Above, with an empty line between the items
vsep :: [Doc] -> Doc
vsep docs = vcat (map (\d -> d $$ (text "")) docs)

-- Like punctuate, but places a separator at the end, too
punctuate2 p []     = []
punctuate2 p (d:ds) = go d ds
                   where
                     go d [] = [d] ++ [p]
                     go d (e:es) = (d <> p) : go e es

string_split :: String -> Char -> [String]
string_split input c = f input [] []
	where
	f :: String	-> [String] -> String -> [String]
	f [] result part = result ++ [part]
	f (x:xs) result part
		| x == c
		= f xs (result ++ [part]) []
		| otherwise
		= f xs result (part ++ [x])

string_replace_char :: String -> Char -> Char -> String
string_replace_char input from_c to_c = map f input
	where
	f c = if (c == from_c) then to_c else c

string_capitalize_first_char :: String -> String
string_capitalize_first_char input = f input
	where
	f [] = []
	f (x:xs) = (toUpper x:xs)

-- Generate the Haskell name from the CORBA Repository ID
-- It is assumed that the repository id is in OMG IDL format.
-- Haskell types must begin with an uppercase letter, while function names
-- must begin with a lowercase letter. This is achieved by pretending the
-- string 'C_' to type names, and the 'c_' string to function names

gen_haskell_name :: String -> String
gen_haskell_name "IDL:CORBA/Object:1.0"
	= "CORBA_Object"
gen_haskell_name "IDL:omg.org/CORBA/Object:1.0" 
	= "CORBA_Object"
gen_haskell_name repoid 
	= "C_" ++ ((string_replace_char (remove_omg_prefix ((string_split repoid ':')!!1))) '/' '_')

to_function_name :: String -> String
to_function_name ('C':('_':xs)) = "c_" ++ xs
to_function_name xs = if (isPrefixOf "CORBA_" xs) then ("corba_" ++ (drop 6 xs)) else xs

-- Remove the 'omg.org' prefix, it it exists
remove_omg_prefix :: String -> String
remove_omg_prefix repoid = f (splitAt 8 repoid)
	where
	f (['o', 'm', 'g', '.', 'o', 'r', 'g', '/'], rest) = rest
	f (head, tail) = head ++ tail

-- Convert the list into a string, inserting a delimiter between elements
toStringWith :: [String] -> String -> String
toStringWith [] _	=	""
toStringWith [x] _	=	x
toStringWith (x:xs)	delim =	x ++ delim ++ (toStringWith xs delim)

-- Convert the list into a string, inserting a delimiter before elements
toStringWith2 :: [String] -> String -> String
toStringWith2 []     _		=	""
toStringWith2 [x]    delim	=	delim ++ x
toStringWith2 (x:xs) delim 	=	delim ++ x ++ (toStringWith2 xs delim)

stringify :: String -> String
stringify s = "\"" ++ s ++ "\""

charify :: Char -> String
charify c = "'" ++ [c] ++ "'"

-- Replace the '%' symbols in arg1 with arg2
string_subst :: String -> String -> String
string_subst s1 s2 = foldr (++) "" (map f s1)
	where
	f '%'	=	s2
	f c		=	[c]

-- Replace the '%' symbols in arg1 with arg2, and the '$' symbols with arg3
string_subst2 :: String -> String -> String -> String
string_subst2 s1 s2 s3 = foldr (++) "" (map f s1)
	where
	f '%'	=	s2
	f '$'   =	s3
	f c		=	[c]

-- Answer the Haskell type name for the given TypeCode.
haskell_type_for :: TypeCode -> String
haskell_type_for tc = case tc of
 	Tk_Void			->	"CORBA_Void"
 	Tk_Short		->	"CORBA_Short"
 	Tk_Long			->	"CORBA_Long"
 	Tk_UShort		->	"CORBA_UShort"
	Tk_ULong		->	"CORBA_ULong"
	Tk_Float		->	"CORBA_Float"
	Tk_Double		->	"CORBA_Real"
	Tk_Boolean		->	"CORBA_Boolean"
	Tk_Char			->	"CORBA_Char"
	Tk_Octet		->	"CORBA_Octet"
	Tk_Any			->	"CORBA_Any"
	(Tk_String _)	->	"CORBA_String"
	Tk_TypeCode		->	"CORBA_TypeCode"
	(Tk_ObjRef repoid _)	->	gen_haskell_name repoid
	(Tk_Struct repoid _ _) 	->	gen_haskell_name repoid
	(Tk_Union repoid _ _ _)	->	gen_haskell_name repoid
	(Tk_Enum repoid _ _)	->	gen_haskell_name repoid
	(Tk_Alias repoid _ _)	->	gen_haskell_name repoid
	(Tk_Exception repoid _ _)	->	gen_haskell_name repoid
	(Tk_Recursive repoid)	->	gen_haskell_name repoid
	tc				->	error ("haskell_type_for called with" ++ (show tc))

-- Same as above, but handle anonymous types too
typename_for :: TypeCode -> String -> String
typename_for tc name
	= if (is_anonym_member tc) then name else (haskell_type_for tc)

-- Answer a Haskell expression representing this TypeCode
-- The second argument is used to name anonymous types
typecode_for :: TypeCode -> String -> String
typecode_for (Tk_String bound) _ 
	= "(Tk_String " ++ (show bound) ++ ")"
typecode_for tc name 
	= to_function_name $ (typename_for tc name) ++ "__tc"

typecode_for_name :: String -> Doc
typecode_for_name name = text $ to_function_name (name ++ "__tc")

-- Answer the name of the decoder function of this type
decode_for :: TypeCode -> Doc
decode_for tc = text $ to_function_name $ (haskell_type_for tc) ++ "__decode"

decode_for_name :: String -> Doc
decode_for_name name = text $ to_function_name (name ++ "__decode")

-- Answer the name of the encoder function of this type
encode_for :: TypeCode -> Doc
encode_for tc = text $ to_function_name $ (haskell_type_for tc) ++ "__encode"

encode_for_name :: String -> Doc
encode_for_name name = text $ to_function_name (name ++ "__encode")

-- Answer the name of the function returning a default value of this type
-- This is used in constructing OUT arguments
default_value_for :: TypeCode -> Doc
default_value_for tc = text $ to_function_name $ (haskell_type_for tc) ++ "__default"

haskell_type_for_anonym_member :: String -> String -> String
haskell_type_for_anonym_member parent_name member_name = parent_name ++ "__members__" ++ member_name

-- Answer whenever this typecode denotes an anonymous type
is_anonym_member :: TypeCode -> Bool
is_anonym_member (Tk_Sequence _ _) 	=	True
is_anonym_member (Tk_Array _ _) 	=	True
is_anonym_member _					=	False

haskell_value_for any = case any of
	(CorbaShort i)	->	show i
	(CorbaLong i)	->	show i
	(CorbaUShort i)	->	show i
	(CorbaULong i)	->	show i
	(CorbaFloat f)	->	show f
	(CorbaDouble f)	->	show f
	(CorbaBoolean b)->	show b
	(CorbaChar c)	->	charify c
	(CorbaOctet i)	->	show i
	(CorbaString s)	->	stringify s
	(CorbaEnum tc@(Tk_Enum repoid _ _) s)-> string_capitalize_first_char s

-- Return a Haskell expression representing this Any
haskell_encoding_for any = case any of
	(CorbaShort i)	->	"(CorbaShort " ++ (show i) ++ ")"
	(CorbaLong i)	->	"(CorbaLong " ++ (show i) ++ ")"
	(CorbaUShort i)	->	"(CorbaUShort " ++ (show i) ++ ")"
	(CorbaULong i)	->	"(CorbaULong " ++ (show i) ++ ")"
	(CorbaFloat f)	->	"(CorbaFloat " ++ (show f) ++ ")"
	(CorbaDouble f)	->	"(CorbaDouble " ++ (show f) ++ ")"
	(CorbaBoolean b)->	"(CorbaBoolean " ++ (show b) ++ ")"
	(CorbaChar c)	->	"(CorbaChar " ++ (show c) ++ ")"
	(CorbaOctet i)	->	"(CorbaOctet " ++ (show i) ++ ")"
	(CorbaEnum tc@(Tk_Enum repoid _ _) s)-> "(CorbaEnum " ++ (typecode_for tc "") ++ " " ++ s ++ ")"

-- Generate a header comment for a type
gen_header_comment :: Doc
gen_header_comment = (text $ replicate 32 '-') 	$$ 
					 (text "--     %") 			$$
					 (text $ replicate 32 '-')

-- Generate the signature of the decode function for a type
gen_decode_def :: Doc
gen_decode_def = text "$__decode :: Any -> IO (Maybe %)"

-- Generate the signature of the encode function for a type
gen_encode_def :: Doc
gen_encode_def	= text "$__encode :: % -> Any"

-- Generate the signature of the typecode function for a type
gen_tc_def :: Doc
gen_tc_def = text "$__tc :: TypeCode"

-- Answer the pair of get/set operations for the given attribute
get_operations_for :: C_CORBA_AttributeDescription -> C_CORBA_IDLType -> [C_CORBA_OperationDescription]
get_operations_for attr idltype_nil
	= [read_desc] ++ (if read_write then [write_desc] else [])
	where
		(C_CORBA_AttributeDescription name _ defined_in _ type2 mode) = attr

		read_desc = (C_CORBA_OperationDescription
			("_get_" ++ name)		--	name
			"foo"					-- 	id (A bit ugly...)
			defined_in				--	defined_in
			"1.0"					--	version
			type2					--	result
			OP_NORMAL				--	mode
			[]						--	contexts
			[]						--	parameters
			[]						--	exceptions
			)

		read_write	= (mode == ATTR_NORMAL)

		write_desc = (C_CORBA_OperationDescription
			("_set_" ++ name)		--	name
			"foo"					-- 	id (A bit ugly...)
			defined_in				--	defined_in
			"1.0"					--	version
			Tk_Void					--	result
			OP_NORMAL				--	mode
			[]						--	contexts
			[write_arg]				--	parameters
			[]						--	exceptions
			)

		write_arg = (C_CORBA_ParameterDescription
			"attr"					--	name
			type2					--	type
			-- This field contains an object reference, but I need an IO action
			--  to create one. Hence this hack...
			idltype_nil				--	type_def
			PARAM_IN				--	mode
			)

--------------------------------
--      IDL COMPILER
--------------------------------

data CompilerState  = CompilerState {	
	-- The 'compile_attribute' function
	-- needs an object of this type, hence this hack...
	idlType_nil 	:: C_CORBA_IDLType,
	-- Determines whenever server side bindings are generated
	genServerSide	:: Bool,
	-- Only used by compile_operation. Signals that server side
	-- operation definitions should be generated
	genServantDef	:: Bool,
	-- Only used by compile_operation. Signals that server side 
	-- wrapper functions should be generated
	genWrapper		:: Bool
					 }

compile_ir :: CompilerState -> IRObject -> String
compile_ir opt ir = case ir of
	(IREnum desc)				->	compile_enum opt desc
	(IRStruct desc members)		->	compile_struct opt desc members
	(IRRepository members) 		->	compile_repository opt members
	(IRInterface desc members) 	->	compile_interface opt desc members
	(IRModule desc members) 	->	compile_module opt desc members
	(IRConstant desc) 			->	compile_constant opt desc
	(IRException desc) 			->	compile_exception opt desc
	(IRUnion desc members) 		->	compile_union opt desc members
	(IRAlias desc) 				->	compile_alias opt desc
	(IROperation desc) 			->	compile_operation opt desc
	(IRAttribute desc) 			->	compile_attribute opt desc

--------------------------------
--    ANONYM TYPES
--------------------------------

compile_anonym_type :: CompilerState -> String -> TypeCode -> String
compile_anonym_type opt name tc@(Tk_Sequence _ _) 
	= compile_sequence opt name tc
compile_anonym_type opt name tc@(Tk_Array _ _) 
	= compile_array opt name tc

--------------------------------
--        SEQUENCES
--------------------------------

compile_sequence :: CompilerState -> String -> TypeCode -> String
compile_sequence opt name (Tk_Sequence length tc)
	= gen_member_type ++ (string_subst2 (render (vcat components)) haskell_name fun_name)
		where
		components 
			= [gen_header_comment, gen_def, gen_tc_def, 
			   gen_tc_impl, gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type
		-- The '$' symbol is substituted for the function name prefix

		gen_member_type = if (is_anonym_member tc) then ((compile_anonym_type opt member_type_name tc) ++ "\n") else ""

		gen_def = (text "type % = [") <> (text member_type_name) <> (text "]")
		
		gen_tc_impl = 
			(text "$__tc = Tk_Sequence") <+> (int length) <+> (text "(") <+> (text (typecode_for tc member_type_name)) <+> (text ")")

		gen_decode_impl	=
			text "$__decode any = corba_seq_decode any" <+> (text $ typecode_for tc member_type_name) <+> (decode_for_name member_type_name)
						  
		gen_encode_impl	= (text "$__encode value = (CorbaSequence $__tc (map")  <+> (encode_for_name member_type_name) <+> (text " value))")

		-- Utility functions

		haskell_name = name

		fun_name	 = to_function_name haskell_name

		member_type_name = if (is_anonym_member tc) then (haskell_name ++ "__member_type") else (haskell_type_for tc)
		
--------------------------------
--         ARRAYS
--------------------------------

compile_array :: CompilerState -> String -> TypeCode -> String
compile_array opt name (Tk_Array length tc)
	= gen_member_type ++ (string_subst2 (render (vcat components)) haskell_name fun_name)
		where
		components 
			= [gen_header_comment, gen_def, 
			   gen_tc_def, gen_tc_impl, 
			   gen_decode_def, gen_decode_impl, gen_encode_def, gen_encode_impl]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type
		-- The '$' symbol is substituted for the function name prefix

		gen_member_type = if (is_anonym_member tc) then ((compile_anonym_type opt member_type_name tc) ++ "\n") else ""

		gen_def = text "type % = [" <> text member_type_name <> text "]"
		
		gen_tc_impl = 
			text "$__tc = Tk_Array" <+> (int length) <+> (parens (text (typecode_for tc member_type_name)))

		gen_decode_impl	= 
			text "$__decode any = corba_seq_decode any" <+> (text $ typecode_for tc member_type_name) <+> (decode_for_name member_type_name)

		gen_encode_impl	= text "$__encode value = (CorbaArray $__tc (map " <+> (encode_for_name member_type_name) <+> text " value))"

		-- Utility functions

		haskell_name = name
		fun_name	 = to_function_name haskell_name
		member_type_name = if (is_anonym_member tc) then (haskell_name ++ "__member_type") else (haskell_type_for tc)
		
--------------------------------
--       REPOSITORIES
--------------------------------

compile_repository opt members
	=  render $ hcat $ (punctuate (text "\n\n") (map (\o -> text $ compile_ir opt o) members))

--------------------------------
--        ENUMS
--------------------------------

compile_enum opt desc
	= string_subst2 (render (vcat components)) haskell_name fun_name
		where
		components 
			= [gen_header_comment,
			   gen_def, gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_names, gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		-- Code Generation Functions
		-- The '%' symbol is substituted for the name of the haskell type
		-- The '$' symbol is substituted for the function name prefix

		gen_def = text $ "data % = " ++ (toStringWith values "\n\t| ") ++ "\n\t deriving (Eq, Show)"
		
		gen_tc_impl = 
			text "$__tc = Tk_Enum $__repoid $__name $__memberNames"

		gen_repoid 	= text $ "$__repoid \t=\t " ++ (stringify repoid)
		gen_name	= text $ "$__name \t=\t " ++ (stringify name)

		gen_names	= text $ "$__memberNames = [\n\t" ++ (toStringWith (map stringify names) "\n\t,") ++ "\n\t]"

		gen_decode_impl	= 
			text "$__decode any = " 							$$
			text "\tcorba_enum_decode any $__tc decode_value" 	$$
			text "\t	where"									$$
			text (toStringWith (map gen_member names) "\n") 	$$
			text "\t\tdecode_value _ \t=\t Nothing"
			where
			gen_member name = "\t\tdecode_value " ++ (stringify name) ++ "\t=\t(Just " ++ (string_capitalize_first_char name) ++ ")"

		gen_encode_impl	= text $ "$__encode value = (CorbaEnum $__tc (encode_value value))\n	where\n" ++ (concat (map (\n -> ("\tencode_value " ++ (string_capitalize_first_char n) ++ "\t=\t" ++ (stringify n) ++ "\n")) names)) ++ "\n"

		-- Utility Functions

		(C_CORBA_TypeDescription name repoid _ _ type2) = desc

		haskell_name = (gen_haskell_name repoid)

		fun_name = to_function_name haskell_name

		names = get_names type2
			where
			get_names (Tk_Enum repoid name members) = members

		values = map string_capitalize_first_char names
		
--------------------------------
--         STRUCTS
--------------------------------

compile_struct opt desc included_defs
	= (string_subst2 (render (vcat components)) haskell_name fun_name) ++ gen_included_defs
		where
		components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type
		-- The '$' symbol is substituted for the function name prefix

		gen_included_defs = concat (map (\o -> compile_ir opt o) included_defs)

		gen_anonym_members = vcat (map gen_member (filter (is_anonym_member . snd) members))
			where
			gen_member (n, tc) = text $ compile_anonym_type opt (haskell_type_for_anonym_member haskell_name n) tc

		gen_def = text $ "\ndata % = %\n\t" ++ (gen_members members) ++ "\n\tderiving (Show)\n"
			where
			gen_members []	= "Int  -- Dummy"
			gen_members l 	= toStringWith (map gen_member l) "\n\t"
			gen_member (n, tc) = (type_for_member n tc) ++ "\t--\t" ++ n

		gen_tc_impl = 
			text $ "$__tc = Tk_Struct $__repoid $__name " ++ "[\n\t" ++ (toStringWith (map gen_member members) ",\n\t") ++ "\n\t]" ++ "\n"
			where
			gen_member (name, tc) = "(" ++ (stringify name) ++ ", " ++ (typecode_for_member name tc) ++ ")"

		gen_repoid 	= text $ "$__repoid \t=\t \"" ++ repoid ++ "\"\n"
		gen_name	= text $ "$__name \t=\t \"" ++ type_name ++ "\"\n"

		gen_decode_impl = text $
			"$__decode any = corba_struct_decode any $__tc decode_fun\n" ++
			"\t	where\n"									++
			"\t\tdecode_fun members = do\n"						++
			"\t\t" ++ (toStringWith (map gen_member1 (with_indexes members)) "\n\t\t") ++ "\n"							++
			"\t\treturn (if and [" ++ (toStringWith (map (\i -> "is_just m" ++ (show i)) ([0.. length members - 1])) ", ") ++ "] then (Just (%\n" 						++
			"\t\t\t" ++ (gen_members members) ++ "\n\t\t\t)) else Nothing)\n"
			where
			gen_member1 ((name, tc), index) = "m" ++ (show index) ++ " <- " ++ (decode_for_member name tc) ++ " (snd (members!!" ++ (show index) ++ ")) "
			gen_members [] = "0"
			gen_members l = toStringWith (map gen_member (with_indexes members)) "\n\t\t\t"
			gen_member ((name, tc), index) = "\t(just m" ++ (show index) ++ ")"

		gen_encode_impl	= text $ "$__encode (% " ++ (gen_fieldnames) ++ ") = (CorbaStruct $__tc [\n\t" ++ (toStringWith (map gen_member (with_indexes members)) ",\n\t") ++ "\n\t])\n"
			where
			gen_fieldnames 	
				= toStringWith (map gen_fieldname (with_indexes members)) " "
			gen_fieldname ((name, tc), index)
				= "m" ++ (show index)
			gen_member ((name, tc), index)
				= "(" ++ (stringify name) ++ ", " ++ (encode_for_member name tc) ++ " m" ++ (show index) ++ ")"

		-- Utilities

		(C_CORBA_TypeDescription type_name repoid _ _ type2) = desc

		haskell_name = gen_haskell_name repoid

		fun_name = to_function_name haskell_name

		members = get_names type2
			where
			get_names (Tk_Struct repoid name members) = members

		-- Handle anonymous members
		type_for_member :: String -> TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) then (haskell_type_for_anonym_member haskell_name name) else (haskell_type_for tc)

		-- Handle anonymous members
		typecode_for_member :: String -> TypeCode -> String
		typecode_for_member name tc = typecode_for tc (haskell_type_for_anonym_member haskell_name name)

		-- Handle anonymous members
		decode_for_member :: String -> TypeCode -> String
		decode_for_member name tc = render $ if (is_anonym_member tc) then (decode_for_name (haskell_type_for_anonym_member haskell_name name)) else (decode_for tc)

		-- Handle anonymous members
		encode_for_member :: String -> TypeCode -> String
		encode_for_member name tc = render $ if (is_anonym_member tc) then (encode_for_name (haskell_type_for_anonym_member haskell_name name)) else (encode_for tc)

--------------------------------
--         UNIONS
--------------------------------

compile_union opt desc included_defs
	= (string_subst2 (render (vcat components)) haskell_name fun_name) ++ gen_included_defs
		where
		components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type

		gen_included_defs = concat (map (\o -> compile_ir opt o) included_defs)

		gen_anonym_members = text $ toStringWith (map gen_member (filter (\(n,d,t) -> is_anonym_member t) members)) "\n"
			where
			gen_member (n, disc, tc) = compile_anonym_type opt (haskell_type_for_anonym_member haskell_name n) tc

		gen_def = text $ "data % = \n\t" ++ (toStringWith (map gen_member cooked_members) "\n\t| ") ++ "\n\n"
			where
			gen_member member@(n, disc, tc)
				| is_implicit_member member
				= (name_for_discriminator disc) ++ " " ++ (haskell_type_for disc_type)
				| is_default_member (n, disc, tc)
				= (name_for_discriminator disc) ++ " " ++ (haskell_type_for disc_type) ++ " " ++ (type_for_member n tc)
				| otherwise
				= (name_for_discriminator disc) ++ " " ++ (type_for_member n tc)

		gen_tc_impl = text $
			"$__tc = Tk_Union $__repoid $__name " ++ (typecode_for disc_type "") ++ " [\n\t" ++ (toStringWith (map gen_member members) ",\n\t") ++ "\n\t]" ++ "\n"
			where
			gen_member (name, disc, tc) = "(" ++ (stringify name) ++ ", " ++ (haskell_encoding_for disc) ++ ", " ++ (typecode_for_member name tc) ++ ")"

		gen_repoid 	= text $ "$__repoid \t=\t \"" ++ repoid ++ "\"\n"
		gen_name	= text $ "$__name \t=\t \"" ++ type_name ++ "\"\n"

		gen_decode_impl	= text $
			"$__decode any = corba_union_decode any $__tc decode_fun\n" ++
			"\t	where\n"												++
			"\t\t" ++ (toStringWith (map gen_member cooked_members) "\n\t\t") ++ "\n"
			where
			gen_member member@(name, disc, tc)
				-- The cooked_members function guarantees that
				-- the default member is generated last
				| is_implicit_member member =
				"decode_fun disc value = do\n"						++
				"\t\t\t(d) <- " ++ (render $ decode_for_name (haskell_type_for disc_type)) ++ " disc \n"															++
				"\t\t\tif is_just d then return (Just (" ++ (name_for_discriminator disc) ++ " (just d))) else return Nothing\n"
				| is_default_member member =
				"decode_fun disc value = do \n"						++
				"\t\t\t(d) <- " ++ (render $ decode_for_name (haskell_type_for disc_type)) ++ " disc \n"															++
				"\t\t\t(v) <- " ++ (decode_for_member name tc) ++ " value \n"																++
				"\t\t\tif (is_just v && is_just d) then return (Just (" ++ (name_for_discriminator disc) ++ " (just d) (just v))) else return Nothing\n"
				| otherwise =
				"decode_fun " ++ (haskell_encoding_for disc) ++ " value = do \n"	++
				"\t\t\t(v) <-" ++ (decode_for_member name tc) ++ " value \n" ++
				"\t\t\tif is_just v then return (Just (" ++ (name_for_discriminator disc) ++ " (just v))) else return Nothing\n"

		gen_encode_impl	= text $ "$__encode rec = encode rec\n	where\n\t" ++ (toStringWith (map gen_member cooked_members) "\t") ++ "\n"
			where
			gen_member member@(name, disc, tc)
				| is_implicit_member member
				= "encode (" ++ (name_for_discriminator disc) ++ " disc) = (CorbaUnion $__tc (" ++ (render $ encode_for_name (haskell_type_for disc_type)) ++ " disc) (CorbaShort 0))\n"
				| is_default_member member
				= "encode (" ++ (name_for_discriminator disc) ++ " disc value) = (CorbaUnion " ++ (render $ typecode_for_name haskell_name) ++ " (" ++ (render $ (encode_for_name (haskell_type_for disc_type))) ++ " disc) (" ++ (encode_for_member name tc) ++ " value))\n"
				| otherwise
				= "encode (" ++ (name_for_discriminator disc) ++ " value) = (CorbaUnion " ++ (render $ typecode_for_name haskell_name) ++ " " ++ (haskell_encoding_for disc) ++ " (" ++ (encode_for_member name tc) ++ " value))\n"

		-- Utilities

		(C_CORBA_TypeDescription type_name repoid _ _ type2) = desc

		haskell_name = gen_haskell_name repoid

		fun_name = to_function_name haskell_name

		(disc_type, members) = get_info type2
			where
			get_info (Tk_Union repoid name tc members) = (tc, members)

		--
		-- Construct a new list of members as follows:
		--  - if there is a default member, it is placed at the end of the list
		--  - if there is no default member, an implicit default member is added
		--    to the end of the list
		--

		cooked_members
			| not has_default_member
			= members ++ [("foo", (CorbaOctet 0), Tk_Null)]
			| otherwise
			= 
			(filter (not . is_default_member) members) ++
			(filter is_default_member members)

		name_for_discriminator disc = type_name ++ "_" ++ disc_name
			where 
			disc_name = case disc of
				(CorbaShort i) 		-> show i
				(CorbaLong i)  		-> show i
				(CorbaUShort i) 	-> show i
				(CorbaULong i)		-> show i
				(CorbaChar i)		-> error "union discriminator type Char is not supported."
				(CorbaOctet i)		-> "default"
				(CorbaBoolean b)	-> show b
				(CorbaEnum tc s)	-> s

		is_default_member (_, (CorbaOctet i), _) 	= True
		is_default_member _				 			= False

		is_implicit_member :: (String, Any, TypeCode) -> Bool
		is_implicit_member (_, (CorbaOctet i), Tk_Null)	=	True
		is_implicit_member _								=	False

		has_default_member = (length (filter is_default_member members)) > 0

		-- Handle anonymous members
		type_for_member :: String -> TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) then (haskell_type_for_anonym_member haskell_name name) else (haskell_type_for tc)

		-- Handle anonymous members
		typecode_for_member :: String -> TypeCode -> String
		typecode_for_member name tc = typecode_for tc (haskell_type_for_anonym_member haskell_name name)

		-- Handle anonymous members
		decode_for_member :: String -> TypeCode -> String
		decode_for_member name tc = render $ if (is_anonym_member tc) then (decode_for_name (haskell_type_for_anonym_member haskell_name name)) else (decode_for tc)

		-- Handle anonymous members
		encode_for_member :: String -> TypeCode -> String
		encode_for_member name tc = render $ if (is_anonym_member tc) then (encode_for_name (haskell_type_for_anonym_member haskell_name name)) else (encode_for tc)

--------------------------------
--        EXCEPTIONS
--------------------------------

compile_exception opt desc
	= (string_subst2 (render (vcat components)) haskell_name fun_name)
		where
		components 
			= [gen_anonym_members, gen_def, 
			   gen_tc_def, gen_tc_impl, gen_repoid, gen_name, 
			   gen_decode_def, gen_decode_impl, 
			   gen_encode_def, gen_encode_impl]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type

		gen_anonym_members = text $ toStringWith (map gen_member (filter (is_anonym_member . snd) members)) "\n"
			where
			gen_member (n, tc) = compile_anonym_type opt (haskell_type_for_anonym_member haskell_name n) tc

		gen_def = text $ "\ndata % = %\n\t" ++ (gen_members members) ++ "\n\n"
			where
			gen_members []	= "Int  -- Dummy"
			gen_members l 	= toStringWith (map gen_member l) "\n\t"
			gen_member (n, tc) = (type_for_member n tc) ++ "\t--\t" ++ n

		gen_tc_impl = text $
			"$__tc = Tk_Exception $__repoid $__name " ++ "[\n\t" ++ (toStringWith (map gen_member members) ",\n\t") ++ "\n\t]" ++ "\n"
			where
			gen_member (name, tc) = "(" ++ (stringify name) ++ ", " ++ (typecode_for_member name tc) ++ ")"

		gen_repoid 	= text $ "$__repoid \t=\t \"" ++ repoid ++ "\"\n"
		gen_name	= text $ "$__name \t=\t \"" ++ type_name ++ "\"\n"

		gen_decode_impl = text $
			"$__decode any = corba_exception_decode any $__tc decode_fun\n" ++
			"\t	where\n"									++
			"\t\tdecode_fun members = do\n"						++
			"\t\t" ++ (toStringWith (map gen_member1 (with_indexes members)) "\n\t\t") ++ "\n"							++
			"\t\treturn (if and [" ++ (toStringWith (map (\i -> "is_just m" ++ (show i)) ([0.. length members - 1])) ", ") ++ "] then (Just (%\n" 						++
			"\t\t\t" ++ (gen_members members) ++ "\n\t\t\t)) else Nothing)\n"
			where
			gen_member1 ((name, tc), index) = "m" ++ (show index) ++ " <- " ++ (decode_for_member name tc) ++ " (snd (members!!" ++ (show index) ++ ")) "
			gen_members [] = "0"
			gen_members l = toStringWith (map gen_member (with_indexes members)) "\n\t\t\t"
			gen_member ((name, tc), index) = "\t(just m" ++ (show index) ++ ")"

		gen_encode_impl	= text $ "$__encode (% " ++ (gen_fieldnames) ++ ") = (CorbaException $__tc [\n\t" ++ (toStringWith (map gen_member (with_indexes members)) ",\n\t") ++ "\n\t])\n"
			where
			gen_fieldnames 	
				= toStringWith (map gen_fieldname (with_indexes members)) " "
			gen_fieldname ((name, tc), index)
				= "m" ++ (show index)
			gen_member ((name, tc), index)
				= "(" ++ (stringify name) ++ ", " ++ (encode_for_member name tc) ++ " m" ++ (show index) ++ ")"

		-- Utilities

		(C_CORBA_ExceptionDescription type_name repoid _ _ type2) = desc

		haskell_name = gen_haskell_name repoid

		fun_name = to_function_name haskell_name

		members = get_names type2
			where
			get_names (Tk_Exception repoid name members) = members

		-- Handle anonymous members
		type_for_member :: String -> TypeCode -> String
		type_for_member name tc = if (is_anonym_member tc) then (haskell_type_for_anonym_member haskell_name name) else (haskell_type_for tc)

		-- Handle anonymous members
		typecode_for_member :: String -> TypeCode -> String
		typecode_for_member name tc = typecode_for tc (haskell_type_for_anonym_member haskell_name name)

		-- Handle anonymous members
		decode_for_member :: String -> TypeCode -> String
		decode_for_member name tc = render $ if (is_anonym_member tc) then (decode_for_name (haskell_type_for_anonym_member haskell_name name)) else (decode_for tc)

		-- Handle anonymous members
		encode_for_member :: String -> TypeCode -> String
		encode_for_member name tc = render $ if (is_anonym_member tc) then (encode_for_name (haskell_type_for_anonym_member haskell_name name)) else (encode_for tc)


--------------------------------
--        TYPE ALIASES
--------------------------------

compile_alias opt (C_CORBA_TypeDescription _ _ _ _ desc_type)
	= case desc_type of

	-- sequences can only be defined using an alias
	(Tk_Alias repoid name tc@(Tk_Sequence _ _)) -> compile_sequence opt (gen_haskell_name repoid) tc

	-- arrays can only be defined using an alias
	(Tk_Alias repoid name tc@(Tk_Array _ _)) -> compile_array opt (gen_haskell_name repoid) tc

	-- a real alias for an already generated type
	(Tk_Alias repoid name tc) -> (string_subst2 (render (vcat components)) haskell_name fun_name)
		where
		components = [	gen_header_comment, gen_def, gen_tc_def, gen_tc_impl,
						gen_repoid, gen_name, gen_decode, gen_encode	]

		-- Code generator functions

		gen_def = text $ "\ntype % = " ++ aliased_name ++ "\n"

		gen_tc_impl = text "$__tc \t=\t " <> (typecode_for_name aliased_name)

		gen_repoid 	= text $ "$__repoid \t=\t \"" ++ repoid ++ "\"\n"
		gen_name	= text $ "$__name \t=\t \"" ++ name ++ "\"\n"

		gen_decode	= text "$__decode = " <> (decode_for_name aliased_name)

		gen_encode	= text "$__encode = " <> (encode_for_name aliased_name)

		-- Utility functions

		haskell_name = gen_haskell_name repoid

		fun_name = to_function_name haskell_name

		aliased_name = (haskell_type_for tc)

--------------------------------
--        MODULES
--------------------------------

compile_module opt desc members 
	=  render $ hcat $ (punctuate (text "\n\n") (map (\o -> text $ compile_ir opt o) members))

--------------------------------
--       CONSTANTS
--------------------------------

compile_constant opt (C_CORBA_ConstantDescription _ desc_id _ _ _ desc_value)
	= haskell_name ++ " = " ++ haskell_value ++ "\n"
	where
	haskell_name = to_function_name $ gen_haskell_name desc_id
	haskell_value = haskell_value_for desc_value

--------------------------------
--      INTERFACES
--------------------------------

compile_interface opt desc members
	= (string_subst2 (render (vsep components)) haskell_name fun_name) ++ gen_members
		where
		components 
			= [gen_def, gen_tc_def, gen_tc_impl, 
			   gen_repoid, gen_name, gen_create_def, gen_create_impl,
			   gen_decode_def, gen_decode_impl, gen_encode_def, 
			   gen_encode_impl, gen_nil_def, gen_nil_impl,
			   gen_narrow_impl, gen_widen, gen_widen_instances_impl] ++ server_side_components

		server_side_components = 
			case (genServerSide opt) of
			True 	->	[gen_servant_def, gen_servant_open_def, gen_servant_open_impl]
			False	->	[]

		-- Code generator functions
		-- The '%' symbol is substituted for the name of the haskell type

		gen_def = text "\nnewtype %  = % CORBA_Object\n\tderiving (Show)\n"

		gen_tc_impl = text "$__tc \t=\t (Tk_ObjRef $__repoid $__name)"

		gen_repoid 	= (text "$__repoid \t=\t") <+> (doubleQuotes (text repoid))
		gen_name	= (text "$__name \t=\t") <+> (doubleQuotes (text name))

		gen_create_def = text "$__create :: CORBA_Object -> %"

		gen_create_impl = text "$__create ref = (% ref)"

		gen_decode_impl	= 
			(text "$__decode (CorbaObjRef ref) = return (Just ($__create ref))") $$
			(text "$__decode _ = return Nothing")

		gen_encode_impl  = text "$__encode (% ref) = (CorbaObjRef ref)"

		gen_nil_def = text "$__nil :: IO (%)"

		gen_nil_impl = text "$__nil = do\n\t(ref) <-corba_Object__nil \n\treturn (% ref)"

		gen_narrow_impl = 
			(text "class %__narrow_class a where") $$
		    (text "\t$__narrow :: a -> IO (Maybe %)") $$
		    (vcat (map gen_instance base_interfaces))
			where
			gen_instance base_repoid =
				(text "instance %__narrow_class") <+> (text base_name) <+> (text "where") $$
				(nest 4 (text "$__narrow obj = narrow_objref obj $__repoid $__create"))
				where
					base_name = (gen_haskell_name base_repoid)		

		-- The overloaded widen operation is defined here, while
		-- instances for derived interfaces are defined with that interface
		gen_widen = (text "class %__widen_class a 	where") $$ 
					(nest 4 (text "$__widen :: a -> %"))

		-- Generate instances of the widening functions for base interfaces
		gen_widen_instances_impl = vcat (map gen_instance base_interfaces)
			where
			gen_instance base_repoid = (text "instance") <+> (text base_name) <> (text "__widen_class % where") $$
				(nest 4 ((text $ to_function_name base_name) <> (text "__widen (% ref) =") <+> (text $ to_function_name base_name) <> (text "__create ref")))
				where
					base_name = gen_haskell_name base_repoid

		gen_members = render (vsep (map text (map (\o -> compile_ir opt o) members)))

		gen_servant_def = 
			if (length operations == 0) 
			   then text $ "data %__servant l = %__servant\n"
			   else text $ "data %__servant l = %__servant {\n\t" ++
				(toStringWith
					(map gen_op_def operations) ",\n\t ") ++ "\n\t }\n"
			where
			gen_op_def op 	= compile_operation (opt {genServantDef = True}) (new_op op)
			new_op (C_CORBA_OperationDescription name id defined_in version result mode contexts parameters exceptions) = (C_CORBA_OperationDescription
				("impl_" ++ (to_function_name haskell_name) ++ "_" ++ name)
				id defined_in version result mode contexts parameters exceptions)

		gen_servant_open_def = text $ "$__servant_open :: (PSt l) -> (%__servant l) -> IO (CORBA_Object, (PSt l))\n"

		gen_servant_open_impl = text $
			"$__servant_open ps servant \n\t = corba_Server_add_servant ps " ++ (stringify repoid) ++ " serv \n	where\n\tserv = CORBA_Servant_Wrapper dispatch_fun\n" ++ gen_dispatch_fun ++ "\n" ++ gen_wrappers ++ "\n"
			where
			gen_dispatch_fun = "\n" ++ gen_dispatch_impl
			gen_dispatch_impl = "\tdispatch_fun ps req = do\n\t\top <- corba_ServerRequest_operation req\n\t\tcase op of\n" ++ (toStringWith2 (map gen_line operations) "\t\t\t") ++ "\n\t\t\t_\t -> error \"Unknown operation name\""
			gen_line (C_CORBA_OperationDescription op_name _ _ _ _ _ _ _ _) = (stringify op_name) ++ "\t-> $__wrap_" ++ op_name ++ " ps servant req\n"

			gen_wrappers = toStringWith (map gen_wrapper operations) "\n"
				where
				gen_wrapper op = compile_operation (opt {genWrapper = True }) op

		-- Utility functions
		(C_CORBA_InterfaceDescription name repoid _ _ base_base_interfaces) = desc

		base_interfaces = base_base_interfaces ++ ["IDL:CORBA/Object:1.0"]

		haskell_name = (gen_haskell_name repoid)

		fun_name = to_function_name haskell_name

		-- All the methods in the interface, plus the get/set methods
		-- of the attributes
		operations = map (\(IROperation desc) -> desc) ops
			where
			ops 		= (filter is_operation members) ++ arg_ops
			arg_ops		= map (\d -> (IROperation d)) (concat (map (\a -> get_operations_for a (idlType_nil opt)) attributes))
			attributes 	= map (\(IRAttribute desc) -> desc) (filter is_attribute members)
			is_operation (IROperation _)	= True
			is_operation _					= False
			is_attribute (IRAttribute _)	= True
			is_attribute _					= False

--------------------------------
--       OPERATIONS
--------------------------------

compile_operation opt desc
	= render (vcat components)
		where
		components
			| (genServantDef opt)
			= servant_def_components
			| (genWrapper opt)
			= wrapper_components
			| otherwise
			= impl_components

		impl_components 		= [gen_def, gen_impl]
		servant_def_components 	= [gen_servant_def]
		wrapper_components		= [gen_wrapper_def, gen_wrapper_impl]

		-- Code Generation routines

		gen_def = (text $ to_function_name interface_name) <> (text "_") <> (text op_name) <+> (text "::") <+> (text interface_name) <+> (text "->") <+> (hsep $ punctuate2 (text " -> ") (map gen_arg input_args)) <+> (text "IO ((ResultOrException") <+> gen_return_type <+> (text "CorbaException))")
			where
			gen_arg arg = text (haskell_type_for (type_of_arg arg))
			gen_return_type 
				= case has_output_args of
				False	->	 text (haskell_type_for return_type)
				True	->	 parens $ (text (haskell_type_for return_type)) <> (text ",") <+> (hsep $ punctuate (text ", ") (map (\arg -> text (haskell_type_for (type_of_arg arg))) output_args))

		gen_impl =
			(text $ to_function_name interface_name) <> text "_" <> text op_name <+> text "(" <> text interface_name <+> text "obj)" <+> (hsep (map (\arg -> (text "arg_") <> text (name_of_arg arg)) input_args)) <+> text "= do" 	$$ 
			nest 4 ((text "(result, exception,") <+> gen_args_variable <> text ") <- corba_invoke obj" <+> doubleQuotes (text op_name) <+> gen_args <+> text (typecode_for return_type "") <+> text "[]")				$$
			nest 4 (gen_decode_return_value)			  $$
			nest 4 (text "return (corba_result_create") <+> gen_return_value <+> (text "exception)") $$
			empty
			where
			gen_args
				| length args == 0
				= (text "[]")
				| otherwise
				=  brackets (hcat $ punctuate (text ", ") (map gen_arg args))

			gen_arg arg
				| mode == PARAM_OUT
				= parens (text (constructor mode)) <+> (doubleQuotes (text name)) <+> (text (typecode_for type2 ""))
				| otherwise
				= parens ((text (constructor mode)) <+> (doubleQuotes (text name)) <+> (parens (encode_for type2 <+> (text "arg_") <> (text name))))
				where
				name = (name_of_arg arg) 
				type2 = (type_of_arg arg)
				mode = (mode_of_arg arg)
				constructor m = case m of
					PARAM_IN	->	"InArg"
					PARAM_OUT	->	"OutArgTemplate"
					PARAM_INOUT	->	"InOutArg"

			gen_args_variable = text $ if has_output_args then "args" else "_"

			gen_decode_return_value 
				= case has_output_args of
				False	->	text "result <-" <+> decode_for return_type <+> text "result"
				True	->	text "result <-" <+> decode_for return_type <+> text "result" $$ (vcat (map gen_arg (zip output_args [0..])))
				where
				gen_arg (arg, index) = (text "(r") <> (int index) <> (text ") <-") <+> (decode_for (type_of_arg arg)) <+> (text "(corba_arg_get_value (args!!") <> (int index) <> (text "))")

			gen_return_value 
				= case has_output_args of
				False	->	parens (text "just result")
				True	->	parens ((text "(just result),") <+> (hcat $ punctuate (text " , ") (map gen_arg (zip output_args [0..]))))
				where
				gen_arg (arg, index) = parens ((text "just r") <> (int index))

		gen_servant_def = text $ op_name ++ " :: (PSt l) " ++ (toStringWith2 (map gen_arg input_args) " -> ") ++ " \n\t\t-> IO ((PSt l), (ResultOrException " ++ gen_return_type ++ " CorbaException))"
			where
			gen_arg arg = (haskell_type_for (type_of_arg arg))
			gen_return_type 
				= case has_output_args of
				False	->	 (haskell_type_for return_type)
				True	->	 "(" ++ (haskell_type_for return_type) ++ ", " ++ (toStringWith (map (\arg -> (haskell_type_for (type_of_arg arg))) output_args) ", ") ++ ")"

		gen_wrapper_def = text $ "\n" ++ (to_function_name interface_name) ++ "__wrap_" ++ op_name ++ " :: (PSt l) -> (" ++ interface_name ++ "__servant l) -> CORBA_ServerRequest -> IO (PSt l)"

		gen_wrapper_impl = text $
			"\n" ++ (to_function_name interface_name) ++ "__wrap_" ++ op_name ++ " ps servant req = do\n" ++
			"\tcorba_ServerRequest_set_arg_types req " ++ gen_arg_types ++ "\n"												++
			gen_decode_args									++
			gen_call										++
			gen_send_result
			where
			gen_arg_types
				| length args == 0
				= "[]"
				| otherwise
				=  " [\n\t\t" ++ (toStringWith (map gen_arg args) ",\n\t\t") ++ "\n\t\t] " 
				where
				gen_arg (C_CORBA_ParameterDescription name type2 _ mode)
					= "(" ++ (constructor mode) ++ " " ++ (stringify name) ++ " " ++ (typecode_for type2 "") ++ ")"
					where
					constructor m = case m of
						PARAM_IN	->	"InArgTemplate"
						PARAM_OUT	->	"OutArgTemplate"
						PARAM_INOUT	->	"InOutArgTemplate"

			gen_decode_args
				| length args == 0
				= ""
				| otherwise
				= (toStringWith (map gen_arg indexed_input_args) "\n") ++ "\n"
				where
				gen_arg ((C_CORBA_ParameterDescription name type2 _ mode), index)
					= string_subst ("\targ%\t<- corba_ServerRequest_get_arg req %\n\targ%\t<- " ++ (render $ decode_for type2) ++ " arg%\n\targ%    <- return (just arg%)") (show index)

			gen_call
				= "\t(ps, result) <- (impl_" ++ (to_function_name interface_name) ++ "_" ++ op_name ++ " servant) ps " ++ gen_args ++ "\n"
				where
				gen_args 
					= toStringWith (map arg_name_for indexed_input_args) " "
			
			gen_send_result
				= "\t" ++ aux_fun_name ++ " req result\n\treturn ps\n\t	where\n" ++ gen_noresult_case ++ gen_result_case ++ gen_exception_case
				where
				gen_noresult_case
					= "\t\t" ++ aux_fun_name ++ " req NoResult = do return ()\n"

				gen_result_case
					= "\t\t" ++ aux_fun_name ++ " req (Result (res" ++ gen_arglist  ++ ")) = do\n" ++ gen_encode_args ++ "\n" ++ gen_store_result ++ "\n"

				gen_exception_case
					= "\t\t" ++ aux_fun_name ++ " req (Exception ex)\n\t\t\t= corba_ServerRequest_set_exception req (corba_exception_encode ex)\n"

				aux_fun_name = (to_function_name interface_name) ++ "__wrapaux_" ++ op_name

				gen_arglist
					| has_output_args
					= (toStringWith2 (map arg_name_for indexed_output_args) ", ")
					| otherwise
					= ""
				gen_encode_args
					= toStringWith2 (map gen_arg indexed_output_args) "\n"
					where
					gen_arg (arg@(C_CORBA_ParameterDescription name type2 _ mode), index)
						= "\t\t\tcorba_ServerRequest_set_arg req " ++ (show index) ++ " (" ++ (render $ encode_for type2) ++ " " ++ (arg_name_for (arg, index)) ++ ")"

				gen_store_result
					| has_return_value
					= "\t\t\tcorba_ServerRequest_set_result req (" ++ (render $ encode_for return_type) ++ " res)"
					| otherwise
					= "\t\t\treturn ()"

		-- Utility routines

		(C_CORBA_OperationDescription 
			desc_name _ desc_defined_in _ desc_result _ _ desc_parameters _) = desc
		interface_name	= (gen_haskell_name desc_defined_in)
		op_name 		= desc_name

		return_type 	= desc_result

		args 			= desc_parameters

		input_args		= filter is_input_arg args

		output_args		= filter is_output_arg args

		is_input_arg (C_CORBA_ParameterDescription _ _ _ mode) 	= mode /= PARAM_OUT
		is_output_arg (C_CORBA_ParameterDescription _ _ _ mode) 	= mode /= PARAM_IN

		name_of_arg (C_CORBA_ParameterDescription name _ _ _) 	= name
		type_of_arg (C_CORBA_ParameterDescription _ type2 _ _) 	= type2
		mode_of_arg (C_CORBA_ParameterDescription _ _ _ mode) 	= mode

		has_output_args = (length output_args) > 0

		indexed_input_args = filter (\(a, i) -> is_input_arg a) (with_indexes args)

		indexed_output_args = filter (\(a, i) -> is_output_arg a) (with_indexes args)

		arg_name_for (arg, index) = "arg" ++ (show index)

		has_return_value = has_return_value2 return_type
			where
			has_return_value2 (Tk_Void) 	= False
			has_return_value2 _				= True

--------------------------------
--       ATTRIBUTES
--------------------------------

compile_attribute opt desc
	-- Convert it into an operation invocation
	= toStringWith (map (compile_operation opt) ops) "\n"
		where
		ops = get_operations_for desc (idlType_nil opt)

--------------------------------
--         FRONT END
--------------------------------

data CompilerOption = 
	GenServerSide			-- whenever generate server side bindings too
	deriving (Eq)

compile_idl :: [CompilerOption] -> String -> IRObject -> C_CORBA_IDLType -> IO ()
compile_idl options modulename ir idlType_nil = do
	out_file	<-	openFile (real_module_name ++ ".hs") WriteMode
	hPutStr out_file file_prefix
	hPutStr out_file (compile_ir state ir)
	hPutStr out_file file_postfix
	hClose out_file
	where
	state = (CompilerState
					idlType_nil						--	idlType_nil
					(elem GenServerSide options)	-- 	genServerSide
					False						 	--	genServantDef
					False							--	genWrapper
		    	)
	real_module_name = (string_capitalize_first_char modulename)
	file_prefix = "\nmodule " ++ real_module_name ++ " where\n\nimport Corba\n\n"
	file_postfix = "\n-- Local Variables:\n-- tab-width: 4\n-- End:\n"
	
		

--------------------------------
--          MAIN
--------------------------------

-- For testing
m = do
	(orb, args)	<- corba_ORB_init ["foo"] 
	repository <- resolve_ir orb 
	putStr "Loading IDL definitions..."
	ir	<- ir_dump repository 	
	idlType_nil <- c_CORBA_IDLType__nil 
	putStr "Generating code..."
	compile_idl [GenServerSide] module_name ir idlType_nil
		where
		repository_addr 	=	"inet:localhost:4141"
		module_name			=	"test"

		resolve_ir orb = do
		obj <- corba_ORB_bind orb "IDL:omg.org/CORBA/Repository:1.0" repository_addr 
		(Just repository) <- c_CORBA_IRObject__narrow (decode_or_die obj)
		return repository
			where
			decode_or_die (Just obj) = obj
			decode_or_die Nothing = error "unable to bind to IR\n"

main = do
	(input_file, module_name, opts) <- parse_args
	putStr $ "Module name: " ++ module_name ++ "\n"
	exitCode <- system "kill `ps | grep ird` > /dev/null 2>&1"
	exitCode <- system "sleep 1"
	putStr "Starting IR demon...\n"
	exitCode <- system "ird -ORBIIOPAddr inet:localhost:4141 &"
	exitCode <- system "sleep 1"
--	(orb, args)	<- corba_ORB_init ["foo", "-ORBIIOPAddr", "inet:localhost:4141", "-ORBDebugLevel", "255"] 
	(orb, args)	<- corba_ORB_init ["foo"]
	repository <- resolve_ir orb
	putStr "Running the MICO IDL compiler...\n"
	exitCode <- system ("idl -I. --no-codegen-c++ --feed-included-defs --feed-ir -ORBIfaceRepoAddr inet:localhost:4141 " ++ input_file)
	if exitCode /= ExitSuccess then
		do
		putStr "Error: Unable to feed IDL file to the Interface Repository.\n"
		exitFailure
		else
		return ""
	putStr "Loading IDL definitions...\n"
	ir	<- ir_dump repository 	
	idlType_nil <- c_CORBA_IDLType__nil 
	putStr "Generating code...\n"
	compile_idl opts module_name ir idlType_nil
	exitCode <- system "kill `ps | grep ird` > /dev/null 2>&1"
	putStr "Done.\n"
	where
		-- (Input file name, module name, options)
		parse_args :: IO (String, String, [CompilerOption])
		parse_args = do
			args <- getArgs
--			args <- return ["repo.idl"]
			(opts, args, errs) <- return $ getOpt RequireOrder options args
			if length errs > 0 then
				do
				putStrLn $ concat errs ++ usageInfo usage options
				exitFailure
				else
				return ""
			if length args /= 1 then 
				do
				putStrLn usage
				exitFailure
				else
				return ""
			input_file <- return $ args !! 0
			if (length input_file < 4) || ((drop ((length input_file) - 4) input_file) /= ".idl") then 
				do
				putStrLn "Error: Input file must have extension .idl."
				exitFailure
				else
				return ""
			-- Strip extension and directory name
			return (input_file, string_capitalize_first_char $ reverse $ takeWhile (/= '/') (drop 4 (reverse input_file)), opts)

		options :: [OptDescr CompilerOption]
		options = 
			[Option ['s'] 	["server-side"] 	(NoArg GenServerSide)	"generate server side bindings too"]

		usage = "Usage: idl2haskell [OPTION...] input_file"

		resolve_ir orb = do
		obj <- corba_ORB_bind orb "IDL:omg.org/CORBA/Repository:1.0" "inet:localhost:4141"
--		obj <- corba_ORB_resolve_initial_references orb "InterfaceRepository"
		(Just repository) <- c_CORBA_IRObject__narrow (decode_or_die obj)
		return repository
			where
			decode_or_die (Just obj) = obj
			decode_or_die Nothing = error "unable to bind to IR\n"




-- Local Variables:
-- tab-width: 4
-- End:

